import os

import pytest

from featuretools import list_primitives, summarize_primitives
from featuretools.primitives import (
    AddNumericScalar,
    Age,
    Count,
    Day,
    Diff,
    GreaterThan,
    Haversine,
    IsFreeEmailDomain,
    IsNull,
    Last,
    Max,
    Mean,
    Min,
    Mode,
    Month,
    MultiplyBoolean,
    NMostCommon,
    NumCharacters,
    NumericLag,
    NumUnique,
    NumWords,
    PercentTrue,
    Skew,
    Std,
    Sum,
    Weekday,
    Year,
    get_aggregation_primitives,
    get_default_aggregation_primitives,
    get_default_transform_primitives,
    get_transform_primitives,
)
from featuretools.primitives.base import PrimitiveBase
from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.primitives.utils import (
    _check_input_types,
    _get_descriptions,
    _get_summary_primitives,
    _get_unique_input_types,
    list_primitive_files,
    load_primitive_from_file,
)
from featuretools.utils.gen_utils import Library


def test_list_primitives_order():
    df = list_primitives()
    all_primitives = get_transform_primitives()
    all_primitives.update(get_aggregation_primitives())

    for name, primitive in all_primitives.items():
        assert name in df["name"].values
        row = df.loc[df["name"] == name].iloc[0]
        actual_desc = _get_descriptions([primitive])[0]
        if actual_desc:
            assert actual_desc == row["description"]
        assert row["dask_compatible"] == (Library.DASK in primitive.compatibility)
        assert row["valid_inputs"] == ", ".join(
            _get_unique_input_types(primitive.input_types),
        )
        expected_return_type = (
            str(primitive.return_type) if primitive.return_type is not None else None
        )
        assert row["return_type"] == expected_return_type

    types = df["type"].values
    assert "aggregation" in types
    assert "transform" in types


def test_valid_input_types():
    actual = _get_unique_input_types(Haversine.input_types)
    assert actual == {"<ColumnSchema (Logical Type = LatLong)>"}
    actual = _get_unique_input_types(MultiplyBoolean.input_types)
    assert actual == {
        "<ColumnSchema (Logical Type = Boolean)>",
        "<ColumnSchema (Logical Type = BooleanNullable)>",
    }
    actual = _get_unique_input_types(Sum.input_types)
    assert actual == {"<ColumnSchema (Semantic Tags = ['numeric'])>"}


def test_descriptions():
    primitives = {
        NumCharacters: "Calculates the number of characters in a given string, including whitespace and punctuation.",
        Day: "Determines the day of the month from a datetime.",
        Last: "Determines the last value in a list.",
        GreaterThan: "Determines if values in one list are greater than another list.",
    }
    assert _get_descriptions(list(primitives.keys())) == list(primitives.values())


def test_get_descriptions_doesnt_truncate_primitive_description():
    # single line
    descr = _get_descriptions([IsNull])
    assert descr[0] == "Determines if a value is null."

    # multiple line; one sentence
    descr = _get_descriptions([Diff])
    assert (
        descr[0]
        == "Computes the difference between the value in a list and the previous value in that list."
    )

    # multiple lines; multiple sentences
    class TestPrimitive(TransformPrimitive):
        """This is text that continues on after the line break
            and ends in a period.
            This is text on one line without a period

        Examples:
            >>> absolute = Absolute()
            >>> absolute([3.0, -5.0, -2.4]).tolist()
            [3.0, 5.0, 2.4]
        """

        name = "test_primitive"

    descr = _get_descriptions([TestPrimitive])
    assert (
        descr[0]
        == "This is text that continues on after the line break and ends in a period. This is text on one line without a period"
    )

    # docstring ends after description
    class TestPrimitive2(TransformPrimitive):
        """This is text that continues on after the line break
        and ends in a period.
        This is text on one line without a period
        """

        name = "test_primitive"

    descr = _get_descriptions([TestPrimitive2])
    assert (
        descr[0]
        == "This is text that continues on after the line break and ends in a period. This is text on one line without a period"
    )


def test_get_default_aggregation_primitives():
    primitives = get_default_aggregation_primitives()
    expected_primitives = [
        Sum,
        Std,
        Max,
        Skew,
        Min,
        Mean,
        Count,
        PercentTrue,
        NumUnique,
        Mode,
    ]
    assert set(primitives) == set(expected_primitives)


def test_get_default_transform_primitives():
    primitives = get_default_transform_primitives()
    expected_primitives = [
        Age,
        Day,
        Year,
        Month,
        Weekday,
        Haversine,
        NumWords,
        NumCharacters,
    ]
    assert set(primitives) == set(expected_primitives)


@pytest.fixture
def this_dir():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def primitives_to_install_dir(this_dir):
    return os.path.join(this_dir, "primitives_to_install")


@pytest.fixture
def bad_primitives_files_dir(this_dir):
    return os.path.join(this_dir, "bad_primitive_files")


def test_list_primitive_files(primitives_to_install_dir):
    files = list_primitive_files(primitives_to_install_dir)
    custom_max_file = os.path.join(primitives_to_install_dir, "custom_max.py")
    custom_mean_file = os.path.join(primitives_to_install_dir, "custom_mean.py")
    custom_sum_file = os.path.join(primitives_to_install_dir, "custom_sum.py")
    assert {custom_max_file, custom_mean_file, custom_sum_file}.issubset(set(files))


def test_load_primitive_from_file(primitives_to_install_dir):
    primitve_file = os.path.join(primitives_to_install_dir, "custom_max.py")
    primitive_name, primitive_obj = load_primitive_from_file(primitve_file)
    assert issubclass(primitive_obj, PrimitiveBase)


def test_errors_more_than_one_primitive_in_file(bad_primitives_files_dir):
    primitive_file = os.path.join(bad_primitives_files_dir, "multiple_primitives.py")
    error_text = "More than one primitive defined in file {}".format(primitive_file)
    with pytest.raises(RuntimeError) as excinfo:
        load_primitive_from_file(primitive_file)
    assert str(excinfo.value) == error_text


def test_errors_no_primitive_in_file(bad_primitives_files_dir):
    primitive_file = os.path.join(bad_primitives_files_dir, "no_primitives.py")
    error_text = "No primitive defined in file {}".format(primitive_file)
    with pytest.raises(RuntimeError) as excinfo:
        load_primitive_from_file(primitive_file)
    assert str(excinfo.value) == error_text


def test_check_input_types():
    primitives = [Sum, Weekday, PercentTrue, Day, Std, NumericLag]
    log_in_type_checks = set()
    sem_tag_type_checks = set()
    unique_input_types = set()
    expected_log_in_check = {
        "boolean_nullable",
        "boolean",
        "datetime",
    }
    expected_sem_tag_type_check = {"numeric", "time_index"}
    expected_unique_input_types = {
        "<ColumnSchema (Logical Type = BooleanNullable)>",
        "<ColumnSchema (Semantic Tags = ['numeric'])>",
        "<ColumnSchema (Logical Type = Boolean)>",
        "<ColumnSchema (Logical Type = Datetime)>",
        "<ColumnSchema (Semantic Tags = ['time_index'])>",
    }
    for prim in primitives:
        input_types_flattened = prim.flatten_nested_input_types(prim.input_types)
        _check_input_types(
            input_types_flattened,
            log_in_type_checks,
            sem_tag_type_checks,
            unique_input_types,
        )

    assert log_in_type_checks == expected_log_in_check
    assert sem_tag_type_checks == expected_sem_tag_type_check
    assert unique_input_types == expected_unique_input_types


def test_get_summary_primitives():
    primitives = [
        Sum,
        Weekday,
        PercentTrue,
        Day,
        Std,
        NumericLag,
        AddNumericScalar,
        IsFreeEmailDomain,
        NMostCommon,
    ]
    primitives_summary = _get_summary_primitives(primitives)
    expected_unique_input_types = 7
    expected_unique_output_types = 6
    expected_uses_multi_input = 2
    expected_uses_multi_output = 1
    expected_uses_external_data = 1
    expected_controllable = 3
    expected_datetime_inputs = 2
    expected_bool = 1
    expected_bool_nullable = 1
    expected_time_index_tag = 1

    assert (
        primitives_summary["general_metrics"]["unique_input_types"]
        == expected_unique_input_types
    )
    assert (
        primitives_summary["general_metrics"]["unique_output_types"]
        == expected_unique_output_types
    )
    assert (
        primitives_summary["general_metrics"]["uses_multi_input"]
        == expected_uses_multi_input
    )
    assert (
        primitives_summary["general_metrics"]["uses_multi_output"]
        == expected_uses_multi_output
    )
    assert (
        primitives_summary["general_metrics"]["uses_external_data"]
        == expected_uses_external_data
    )
    assert (
        primitives_summary["general_metrics"]["are_controllable"]
        == expected_controllable
    )
    assert (
        primitives_summary["semantic_tag_metrics"]["time_index"]
        == expected_time_index_tag
    )
    assert (
        primitives_summary["logical_type_input_metrics"]["datetime"]
        == expected_datetime_inputs
    )
    assert primitives_summary["logical_type_input_metrics"]["boolean"] == expected_bool
    assert (
        primitives_summary["logical_type_input_metrics"]["boolean_nullable"]
        == expected_bool_nullable
    )


def test_summarize_primitives():
    df = summarize_primitives()
    trans_prims = get_transform_primitives()
    agg_prims = get_aggregation_primitives()
    tot_trans = len(trans_prims)
    tot_agg = len(agg_prims)
    tot_prims = tot_trans + tot_agg

    assert df["Count"].iloc[0] == tot_prims
    assert df["Count"].iloc[1] == tot_agg
    assert df["Count"].iloc[2] == tot_trans



import pandas as pd
import pytest
from woodwork.logical_types import (
    Datetime,
    Double,
    Integer,
    IntegerNullable,
    NaturalLanguage,
)

from featuretools import dfs
from featuretools.entityset import EntitySet
from featuretools.utils.gen_utils import import_or_none

dd = import_or_none("dask.dataframe")


@pytest.mark.skipif("not dd")
def test_single_table_dask_entityset():
    primitives_list = [
        "absolute",
        "is_weekend",
        "year",
        "day",
        "num_characters",
        "num_words",
    ]

    dask_es = EntitySet(id="dask_es")
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "values": [1, 12, -34, 27],
            "dates": [
                pd.to_datetime("2019-01-10"),
                pd.to_datetime("2019-02-03"),
                pd.to_datetime("2019-01-01"),
                pd.to_datetime("2017-08-25"),
            ],
            "strings": ["I am a string", "23", "abcdef ghijk", ""],
        },
    )
    values_dd = dd.from_pandas(df, npartitions=2)
    ltypes = {"values": Integer, "dates": Datetime, "strings": NaturalLanguage}
    dask_es.add_dataframe(
        dataframe_name="data",
        dataframe=values_dd,
        index="id",
        logical_types=ltypes,
    )

    dask_fm, _ = dfs(
        entityset=dask_es,
        target_dataframe_name="data",
        trans_primitives=primitives_list,
    )

    pd_es = EntitySet(id="pd_es")
    pd_es.add_dataframe(
        dataframe_name="data",
        dataframe=df,
        index="id",
        logical_types=ltypes,
    )

    fm, _ = dfs(
        entityset=pd_es,
        target_dataframe_name="data",
        trans_primitives=primitives_list,
    )

    # Use the same columns and make sure both indexes are sorted the same
    # update the type of the future index column so it doesn't conflict with the pandas fm
    dask_fm = dask_fm.compute().astype({"id": "int64"})
    dask_computed_fm = dask_fm.set_index("id").loc[fm.index][fm.columns]
    pd.testing.assert_frame_equal(fm, dask_computed_fm, check_dtype=False)


@pytest.mark.skipif("not dd")
def test_single_table_dask_entityset_ids_not_sorted():
    primitives_list = [
        "absolute",
        "is_weekend",
        "year",
        "day",
        "num_characters",
        "num_words",
    ]

    dask_es = EntitySet(id="dask_es")
    df = pd.DataFrame(
        {
            "id": [2, 0, 1, 3],
            "values": [1, 12, -34, 27],
            "dates": [
                pd.to_datetime("2019-01-10"),
                pd.to_datetime("2019-02-03"),
                pd.to_datetime("2019-01-01"),
                pd.to_datetime("2017-08-25"),
            ],
            "strings": ["I am a string", "23", "abcdef ghijk", ""],
        },
    )
    values_dd = dd.from_pandas(df, npartitions=2)
    ltypes = {"values": Integer, "dates": Datetime, "strings": NaturalLanguage}
    dask_es.add_dataframe(
        dataframe_name="data",
        dataframe=values_dd,
        index="id",
        logical_types=ltypes,
    )

    dask_fm, _ = dfs(
        entityset=dask_es,
        target_dataframe_name="data",
        trans_primitives=primitives_list,
    )

    pd_es = EntitySet(id="pd_es")
    pd_es.add_dataframe(
        dataframe_name="data",
        dataframe=df,
        index="id",
        logical_types=ltypes,
    )

    fm, _ = dfs(
        entityset=pd_es,
        target_dataframe_name="data",
        trans_primitives=primitives_list,
    )

    # Make sure both indexes are sorted the same
    dask_fm = dask_fm.compute().astype({"id": "int64"})
    pd.testing.assert_frame_equal(
        fm,
        dask_fm.set_index("id").loc[fm.index],
        check_dtype=False,
    )


@pytest.mark.skipif("not dd")
def test_single_table_dask_entityset_with_instance_ids():
    primitives_list = [
        "absolute",
        "is_weekend",
        "year",
        "day",
        "num_characters",
        "num_words",
    ]
    instance_ids = [0, 1, 3]

    dask_es = EntitySet(id="dask_es")
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "values": [1, 12, -34, 27],
            "dates": [
                pd.to_datetime("2019-01-10"),
                pd.to_datetime("2019-02-03"),
                pd.to_datetime("2019-01-01"),
                pd.to_datetime("2017-08-25"),
            ],
            "strings": ["I am a string", "23", "abcdef ghijk", ""],
        },
    )

    values_dd = dd.from_pandas(df, npartitions=2)
    ltypes = {"values": Integer, "dates": Datetime, "strings": NaturalLanguage}
    dask_es.add_dataframe(
        dataframe_name="data",
        dataframe=values_dd,
        index="id",
        logical_types=ltypes,
    )

    dask_fm, _ = dfs(
        entityset=dask_es,
        target_dataframe_name="data",
        trans_primitives=primitives_list,
        instance_ids=instance_ids,
    )

    pd_es = EntitySet(id="pd_es")
    pd_es.add_dataframe(
        dataframe_name="data",
        dataframe=df,
        index="id",
        logical_types=ltypes,
    )

    fm, _ = dfs(
        entityset=pd_es,
        target_dataframe_name="data",
        trans_primitives=primitives_list,
        instance_ids=instance_ids,
    )

    # Make sure both indexes are sorted the same
    dask_fm = dask_fm.compute().astype({"id": "int64"})
    pd.testing.assert_frame_equal(
        fm,
        dask_fm.set_index("id").loc[fm.index],
        check_dtype=False,
    )


@pytest.mark.skipif("not dd")
def test_single_table_dask_entityset_single_cutoff_time():
    primitives_list = [
        "absolute",
        "is_weekend",
        "year",
        "day",
        "num_characters",
        "num_words",
    ]

    dask_es = EntitySet(id="dask_es")
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "values": [1, 12, -34, 27],
            "dates": [
                pd.to_datetime("2019-01-10"),
                pd.to_datetime("2019-02-03"),
                pd.to_datetime("2019-01-01"),
                pd.to_datetime("2017-08-25"),
            ],
            "strings": ["I am a string", "23", "abcdef ghijk", ""],
        },
    )
    values_dd = dd.from_pandas(df, npartitions=2)
    ltypes = {"values": Integer, "dates": Datetime, "strings": NaturalLanguage}
    dask_es.add_dataframe(
        dataframe_name="data",
        dataframe=values_dd,
        index="id",
        logical_types=ltypes,
    )

    dask_fm, _ = dfs(
        entityset=dask_es,
        target_dataframe_name="data",
        trans_primitives=primitives_list,
        cutoff_time=pd.Timestamp("2019-01-05 04:00"),
    )

    pd_es = EntitySet(id="pd_es")
    pd_es.add_dataframe(
        dataframe_name="data",
        dataframe=df,
        index="id",
        logical_types=ltypes,
    )

    fm, _ = dfs(
        entityset=pd_es,
        target_dataframe_name="data",
        trans_primitives=primitives_list,
        cutoff_time=pd.Timestamp("2019-01-05 04:00"),
    )

    # Make sure both indexes are sorted the same
    dask_fm = dask_fm.compute().astype({"id": "int64"})
    pd.testing.assert_frame_equal(
        fm,
        dask_fm.set_index("id").loc[fm.index],
        check_dtype=False,
    )


@pytest.mark.skipif("not dd")
def test_single_table_dask_entityset_cutoff_time_df():
    primitives_list = [
        "absolute",
        "is_weekend",
        "year",
        "day",
        "num_characters",
        "num_words",
    ]

    dask_es = EntitySet(id="dask_es")
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "values": [1, 12, -34],
            "dates": [
                pd.to_datetime("2019-01-10"),
                pd.to_datetime("2019-02-03"),
                pd.to_datetime("2019-01-01"),
            ],
            "strings": ["I am a string", "23", "abcdef ghijk"],
        },
    )
    values_dd = dd.from_pandas(df, npartitions=2)
    ltypes = {"values": IntegerNullable, "dates": Datetime, "strings": NaturalLanguage}
    dask_es.add_dataframe(
        dataframe_name="data",
        dataframe=values_dd,
        index="id",
        time_index="dates",
        logical_types=ltypes,
    )

    ids = [0, 1, 2, 0]
    times = [
        pd.Timestamp("2019-01-05 04:00"),
        pd.Timestamp("2019-01-05 04:00"),
        pd.Timestamp("2019-01-05 04:00"),
        pd.Timestamp("2019-01-15 04:00"),
    ]
    labels = [True, False, True, False]
    cutoff_times = pd.DataFrame(
        {"id": ids, "time": times, "labels": labels},
        columns=["id", "time", "labels"],
    )

    dask_fm, _ = dfs(
        entityset=dask_es,
        target_dataframe_name="data",
        trans_primitives=primitives_list,
        cutoff_time=cutoff_times,
    )

    pd_es = EntitySet(id="pd_es")
    pd_es.add_dataframe(
        dataframe_name="data",
        dataframe=df,
        index="id",
        time_index="dates",
        logical_types=ltypes,
    )

    fm, _ = dfs(
        entityset=pd_es,
        target_dataframe_name="data",
        trans_primitives=primitives_list,
        cutoff_time=cutoff_times,
    )
    # Because row ordering with Dask is not guaranteed, we need to sort on two columns to make sure that values
    # for instance id 0 are compared correctly. Also, make sure the index column has the same dtype.
    fm = fm.sort_values(["id", "labels"])
    dask_fm = dask_fm.compute().astype({"id": "int64"})
    dask_fm = dask_fm.set_index("id").sort_values(["id", "labels"])
    pd.testing.assert_frame_equal(fm, dask_fm, check_dtype=False)


@pytest.mark.skipif("not dd")
def test_single_table_dask_entityset_dates_not_sorted():
    dask_es = EntitySet(id="dask_es")
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "values": [1, 12, -34, 27],
            "dates": [
                pd.to_datetime("2019-01-10"),
                pd.to_datetime("2019-02-03"),
                pd.to_datetime("2019-01-01"),
                pd.to_datetime("2017-08-25"),
            ],
        },
    )

    primitives_list = ["absolute", "is_weekend", "year", "day"]
    values_dd = dd.from_pandas(df, npartitions=1)
    ltypes = {
        "values": Integer,
        "dates": Datetime,
    }
    dask_es.add_dataframe(
        dataframe_name="data",
        dataframe=values_dd,
        index="id",
        time_index="dates",
        logical_types=ltypes,
    )

    dask_fm, _ = dfs(
        entityset=dask_es,
        target_dataframe_name="data",
        trans_primitives=primitives_list,
        max_depth=1,
    )

    pd_es = EntitySet(id="pd_es")
    pd_es.add_dataframe(
        dataframe_name="data",
        dataframe=df,
        index="id",
        time_index="dates",
        logical_types=ltypes,
    )

    fm, _ = dfs(
        entityset=pd_es,
        target_dataframe_name="data",
        trans_primitives=primitives_list,
        max_depth=1,
    )

    dask_fm = dask_fm.compute().astype({"id": "int64"})
    pd.testing.assert_frame_equal(
        fm,
        dask_fm.set_index("id").loc[fm.index],
        check_dtype=False,
    )


@pytest.mark.skipif("not dd")
def test_dask_entityset_secondary_time_index():
    log_df = pd.DataFrame()
    log_df["id"] = [0, 1, 2, 3]
    log_df["scheduled_time"] = pd.to_datetime(
        ["2019-01-01", "2019-01-01", "2019-01-01", "2019-01-01"],
    )
    log_df["departure_time"] = pd.to_datetime(
        [
            "2019-02-01 09:00",
            "2019-02-06 10:00",
            "2019-02-12 10:00",
            "2019-03-01 11:30",
        ],
    )
    log_df["arrival_time"] = pd.to_datetime(
        [
            "2019-02-01 11:23",
            "2019-02-06 12:45",
            "2019-02-12 13:53",
            "2019-03-01 14:07",
        ],
    )
    log_df["delay"] = [-2, 10, 60, 0]
    log_df["flight_id"] = [0, 1, 0, 1]
    log_dask = dd.from_pandas(log_df, npartitions=2)

    flights_df = pd.DataFrame()
    flights_df["id"] = [0, 1, 2, 3]
    flights_df["origin"] = ["BOS", "LAX", "BOS", "LAX"]
    flights_dask = dd.from_pandas(flights_df, npartitions=2)

    pd_es = EntitySet("flights")
    dask_es = EntitySet("flights_dask")

    log_ltypes = {
        "scheduled_time": Datetime,
        "departure_time": Datetime,
        "arrival_time": Datetime,
        "delay": Double,
    }

    pd_es.add_dataframe(
        dataframe_name="logs",
        dataframe=log_df,
        index="id",
        time_index="scheduled_time",
        secondary_time_index={"arrival_time": ["departure_time", "delay"]},
        logical_types=log_ltypes,
    )

    dask_es.add_dataframe(
        dataframe_name="logs",
        dataframe=log_dask,
        index="id",
        logical_types=log_ltypes,
        semantic_tags={"flight_id": "foreign_key"},
        time_index="scheduled_time",
        secondary_time_index={"arrival_time": ["departure_time", "delay"]},
    )

    pd_es.add_dataframe(dataframe_name="flights", dataframe=flights_df, index="id")
    flights_ltypes = pd_es["flights"].ww.logical_types
    dask_es.add_dataframe(
        dataframe_name="flights",
        dataframe=flights_dask,
        index="id",
        logical_types=flights_ltypes,
    )

    pd_es.add_relationship("flights", "id", "logs", "flight_id")
    dask_es.add_relationship("flights", "id", "logs", "flight_id")

    cutoff_df = pd.DataFrame()
    cutoff_df["id"] = [0, 1, 1]
    cutoff_df["time"] = pd.to_datetime(["2019-02-02", "2019-02-02", "2019-02-20"])

    fm, _ = dfs(
        entityset=pd_es,
        target_dataframe_name="logs",
        cutoff_time=cutoff_df,
        agg_primitives=["max"],
        trans_primitives=["month"],
    )

    dask_fm, _ = dfs(
        entityset=dask_es,
        target_dataframe_name="logs",
        cutoff_time=cutoff_df,
        agg_primitives=["max"],
        trans_primitives=["month"],
    )

    # Make sure both matrixes are sorted the same
    # Also need to account for index differences
    dask_fm_computed = dask_fm.compute().astype({"id": "int64"}).set_index("id")
    pd.testing.assert_frame_equal(
        fm.sort_values("delay"),
        dask_fm_computed.sort_values("delay"),
        check_dtype=False,
    )



import numpy as np
import pandas as pd
import pytest
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime

from featuretools.computational_backends.feature_set import FeatureSet
from featuretools.computational_backends.feature_set_calculator import (
    FeatureSetCalculator,
)
from featuretools.feature_base import DirectFeature, Feature, IdentityFeature
from featuretools.primitives import (
    AggregationPrimitive,
    Day,
    Hour,
    Minute,
    Month,
    NMostCommon,
    Second,
    TransformPrimitive,
    Year,
)
from featuretools.primitives.utils import PrimitivesDeserializer
from featuretools.synthesis import dfs
from featuretools.tests.testing_utils import to_pandas
from featuretools.utils.gen_utils import Library


def test_direct_from_identity(es):
    device = Feature(es["sessions"].ww["device_type"])
    d = DirectFeature(base_feature=device, child_dataframe_name="log")

    feature_set = FeatureSet([d])
    calculator = FeatureSetCalculator(es, feature_set=feature_set, time_last=None)
    df = calculator.run(np.array([0, 5]))
    df = to_pandas(df, index="id", sort_index=True)
    v = df[d.get_name()].tolist()
    if es.dataframe_type == Library.SPARK:
        expected = ["0", "1"]
    else:
        expected = [0, 1]
    assert v == expected


def test_direct_from_column(es):
    # should be same behavior as test_direct_from_identity
    device = Feature(es["sessions"].ww["device_type"])
    d = DirectFeature(base_feature=device, child_dataframe_name="log")

    feature_set = FeatureSet([d])
    calculator = FeatureSetCalculator(es, feature_set=feature_set, time_last=None)
    df = calculator.run(np.array([0, 5]))
    df = to_pandas(df, index="id", sort_index=True)
    v = df[d.get_name()].tolist()
    if es.dataframe_type == Library.SPARK:
        expected = ["0", "1"]
    else:
        expected = [0, 1]
    assert v == expected


def test_direct_rename_multioutput(es):
    n_common = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="customers",
        primitive=NMostCommon(n=2),
    )
    feat = DirectFeature(n_common, "sessions")
    copy_feat = feat.rename("session_test")
    assert feat.unique_name() != copy_feat.unique_name()
    assert feat.get_name() != copy_feat.get_name()
    assert (
        feat.base_features[0].generate_name()
        == copy_feat.base_features[0].generate_name()
    )
    assert feat.dataframe_name == copy_feat.dataframe_name


def test_direct_rename(es):
    # should be same behavior as test_direct_from_identity
    feat = DirectFeature(
        base_feature=IdentityFeature(es["sessions"].ww["device_type"]),
        child_dataframe_name="log",
    )
    copy_feat = feat.rename("session_test")
    assert feat.unique_name() != copy_feat.unique_name()
    assert feat.get_name() != copy_feat.get_name()
    assert (
        feat.base_features[0].generate_name()
        == copy_feat.base_features[0].generate_name()
    )
    assert feat.dataframe_name == copy_feat.dataframe_name


def test_direct_copy(games_es):
    home_team = next(
        r for r in games_es.relationships if r._child_column_name == "home_team_id"
    )
    feat = DirectFeature(
        IdentityFeature(games_es["teams"].ww["name"]),
        "games",
        relationship=home_team,
    )
    copied = feat.copy()
    assert copied.dataframe_name == feat.dataframe_name
    assert copied.base_features == feat.base_features
    assert copied.relationship_path == feat.relationship_path


def test_direct_of_multi_output_transform_feat(es):
    # TODO: Update to work with Dask and Spark
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail("Custom primitive is not compatible with Dask or Spark")

    class TestTime(TransformPrimitive):
        name = "test_time"
        input_types = [ColumnSchema(logical_type=Datetime)]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        number_output_features = 6

        def get_function(self):
            def test_f(x):
                times = pd.Series(x)
                units = ["year", "month", "day", "hour", "minute", "second"]
                return [times.apply(lambda x: getattr(x, unit)) for unit in units]

            return test_f

    base_feature = IdentityFeature(es["customers"].ww["signup_date"])
    join_time_split = Feature(base_feature, primitive=TestTime)
    alt_features = [
        Feature(base_feature, primitive=Year),
        Feature(base_feature, primitive=Month),
        Feature(base_feature, primitive=Day),
        Feature(base_feature, primitive=Hour),
        Feature(base_feature, primitive=Minute),
        Feature(base_feature, primitive=Second),
    ]
    fm, fl = dfs(
        entityset=es,
        target_dataframe_name="sessions",
        trans_primitives=[TestTime, Year, Month, Day, Hour, Minute, Second],
    )

    # Get column names of for multi feature and normal features
    subnames = DirectFeature(join_time_split, "sessions").get_feature_names()
    altnames = [DirectFeature(f, "sessions").get_name() for f in alt_features]

    # Check values are equal between
    for col1, col2 in zip(subnames, altnames):
        assert (fm[col1] == fm[col2]).all()


def test_direct_features_of_multi_output_agg_primitives(pd_es):
    class ThreeMostCommonCat(AggregationPrimitive):
        name = "n_most_common_categorical"
        input_types = [ColumnSchema(semantic_tags={"category"})]
        return_type = ColumnSchema(semantic_tags={"category"})
        number_output_features = 3

        def get_function(self, agg_type="pandas"):
            def pd_top3(x):
                counts = x.value_counts()
                counts = counts[counts > 0]
                array = np.array(counts.index[:3])
                if len(array) < 3:
                    filler = np.full(3 - len(array), np.nan)
                    array = np.append(array, filler)
                return array

            return pd_top3

    fm, fl = dfs(
        entityset=pd_es,
        target_dataframe_name="log",
        agg_primitives=[ThreeMostCommonCat],
        trans_primitives=[],
        max_depth=3,
    )

    has_nmost_as_base = []
    for feature in fl:
        is_base = False
        if len(feature.base_features) > 0 and isinstance(
            feature.base_features[0].primitive,
            ThreeMostCommonCat,
        ):
            is_base = True
        has_nmost_as_base.append(is_base)
    assert any(has_nmost_as_base)

    true_result_rows = []
    session_data = {
        0: ["coke zero", "car", np.nan],
        1: ["toothpaste", "brown bag", np.nan],
        2: ["brown bag", np.nan, np.nan],
        3: set(["Haribo sugar-free gummy bears", "coke zero", np.nan]),
        4: ["coke zero", np.nan, np.nan],
        5: ["taco clock", np.nan, np.nan],
    }
    for i, count in enumerate([5, 4, 1, 2, 3, 2]):
        while count > 0:
            true_result_rows.append(session_data[i])
            count -= 1

    tempname = "sessions.N_MOST_COMMON_CATEGORICAL(log.product_id)[%s]"
    for i, row in enumerate(true_result_rows):
        for j in range(3):
            value = fm[tempname % (j)][i]
            if isinstance(row, set):
                assert pd.isnull(value) or value in row
            else:
                assert (pd.isnull(value) and pd.isnull(row[j])) or value == row[j]


def test_direct_with_invalid_init_args(diamond_es):
    customer_to_region = diamond_es.get_forward_relationships("customers")[0]
    error_text = "child_dataframe must be the relationship child dataframe"
    with pytest.raises(AssertionError, match=error_text):
        DirectFeature(
            IdentityFeature(diamond_es["regions"].ww["name"]),
            "stores",
            relationship=customer_to_region,
        )

    transaction_relationships = diamond_es.get_forward_relationships("transactions")
    transaction_to_store = next(
        r for r in transaction_relationships if r.parent_dataframe.ww.name == "stores"
    )
    error_text = "Base feature must be defined on the relationship parent dataframe"
    with pytest.raises(AssertionError, match=error_text):
        DirectFeature(
            IdentityFeature(diamond_es["regions"].ww["name"]),
            "transactions",
            relationship=transaction_to_store,
        )


def test_direct_with_multiple_possible_paths(games_es):
    error_text = (
        "There are multiple relationships to the base dataframe. "
        "You must specify a relationship."
    )
    with pytest.raises(RuntimeError, match=error_text):
        DirectFeature(IdentityFeature(games_es["teams"].ww["name"]), "games")

    # Does not raise if path specified.
    relationship = next(
        r
        for r in games_es.get_forward_relationships("games")
        if r._child_column_name == "home_team_id"
    )
    feat = DirectFeature(
        IdentityFeature(games_es["teams"].ww["name"]),
        "games",
        relationship=relationship,
    )
    assert feat.relationship_path_name() == "teams[home_team_id]"
    assert feat.get_name() == "teams[home_team_id].name"


def test_direct_with_single_possible_path(es):
    feat = DirectFeature(IdentityFeature(es["customers"].ww["age"]), "sessions")
    assert feat.relationship_path_name() == "customers"
    assert feat.get_name() == "customers.age"


def test_direct_with_no_path(diamond_es):
    error_text = 'No relationship from "regions" to "customers" found.'
    with pytest.raises(RuntimeError, match=error_text):
        DirectFeature(IdentityFeature(diamond_es["customers"].ww["name"]), "regions")

    error_text = 'No relationship from "customers" to "customers" found.'
    with pytest.raises(RuntimeError, match=error_text):
        DirectFeature(IdentityFeature(diamond_es["customers"].ww["name"]), "customers")


def test_serialization(es):
    value = IdentityFeature(es["products"].ww["rating"])
    direct = DirectFeature(value, "log")

    log_to_products = next(
        r
        for r in es.get_forward_relationships("log")
        if r.parent_dataframe.ww.name == "products"
    )
    dictionary = {
        "name": direct.get_name(),
        "base_feature": value.unique_name(),
        "relationship": log_to_products.to_dictionary(),
    }

    assert dictionary == direct.get_arguments()
    assert direct == DirectFeature.from_dictionary(
        dictionary,
        es,
        {value.unique_name(): value},
        PrimitivesDeserializer(),
    )



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Boolean, Integer

from featuretools.primitives.base import AggregationPrimitive


class MaxConsecutiveTrue(AggregationPrimitive):
    """Determines the maximum number of consecutive True values in the input

    Examples:
        >>> max_consecutive_true = MaxConsecutiveTrue()
        >>> max_consecutive_true([True, False, True, True, True, False])
        3
    """

    name = "max_consecutive_true"
    input_types = [ColumnSchema(logical_type=Boolean)]
    return_type = ColumnSchema(logical_type=Integer, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = 0

    def get_function(self):
        def max_consecutive_true(x):
            # find the locations where the value changes from the previous value
            not_equal = x != x.shift()
            # use cumulative sum to determine where consecutive values occur. When the
            # sum changes, consecutive False values are present, when the cumulative
            # sum remains unchnaged, consecutive True values are present.
            not_equal_sum = not_equal.cumsum()
            # group the input by the cumulative sum values and use cumulative count
            # to count the number of consecutive values. Add 1 to account for the cumulative
            # sum starting at zero where the first True occurs
            consecutive = x.groupby(not_equal_sum).cumcount() + 1
            # multiply by the original input to keep only the counts that correspond to
            # true values
            consecutive_true = consecutive * x
            # return the max of all the consecutive true values
            return consecutive_true.max()

        return max_consecutive_true



from datetime import datetime
from math import sqrt

import numpy as np
import pandas as pd
import pytest
from pandas.core.dtypes.dtypes import CategoricalDtype
from pytest import raises

from featuretools.primitives import (
    AverageCountPerUnique,
    DateFirstEvent,
    Entropy,
    FirstLastTimeDelta,
    HasNoDuplicates,
    IsMonotonicallyDecreasing,
    IsMonotonicallyIncreasing,
    Kurtosis,
    MaxCount,
    MaxMinDelta,
    MedianCount,
    MinCount,
    NMostCommon,
    NMostCommonFrequency,
    NumFalseSinceLastTrue,
    NumPeaks,
    NumTrueSinceLastFalse,
    NumZeroCrossings,
    NUniqueDays,
    NUniqueDaysOfCalendarYear,
    NUniqueDaysOfMonth,
    NUniqueMonths,
    NUniqueWeeks,
    PercentTrue,
    Trend,
    Variance,
    get_aggregation_primitives,
)
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    check_serialize,
    find_applicable_primitives,
    valid_dfs,
)


def test_nmostcommon_categorical():
    n_most = NMostCommon(3)
    expected = pd.Series([1.0, 2.0, np.nan])

    ints = pd.Series([1, 2, 1, 1]).astype("int64")
    assert pd.Series(n_most(ints)).equals(expected)

    cats = pd.Series([1, 2, 1, 1]).astype("category")
    assert pd.Series(n_most(cats)).equals(expected)

    # Value counts includes data for categories that are not present in data.
    # Make sure these counts are not included in most common outputs
    extra_dtype = CategoricalDtype(categories=[1, 2, 3])
    cats_extra = pd.Series([1, 2, 1, 1]).astype(extra_dtype)
    assert pd.Series(n_most(cats_extra)).equals(expected)


def test_agg_primitives_can_init_without_params():
    agg_primitives = get_aggregation_primitives().values()
    for agg_primitive in agg_primitives:
        agg_primitive()


def test_trend_works_with_different_input_dtypes():
    dates = pd.to_datetime(["2020-01-01", "2020-01-02", "2020-01-03"])
    numeric = pd.Series([1, 2, 3])

    trend = Trend()
    dtypes = ["float64", "int64", "Int64"]

    for dtype in dtypes:
        actual = trend(numeric.astype(dtype), dates)
        assert np.isclose(actual, 1)


def test_percent_true_boolean():
    booleans = pd.Series([True, False, True, pd.NA], dtype="boolean")
    pct_true = PercentTrue()
    pct_true(booleans) == 0.5


class TestAverageCountPerUnique(PrimitiveTestBase):
    primitive = AverageCountPerUnique
    array = pd.Series([1, 1, 2, 2, 3, 4, 5, 6, 7, 8])

    def test_percent_unique(self):
        primitive_func = AverageCountPerUnique().get_function()
        assert primitive_func(self.array) == 1.25

    def test_nans(self):
        primitive_func = AverageCountPerUnique().get_function()
        array_nans = pd.concat([self.array.copy(), pd.Series([np.nan])])
        assert primitive_func(array_nans) == 1.25
        primitive_func = AverageCountPerUnique(skipna=False).get_function()
        array_nans = pd.concat([self.array.copy(), pd.Series([np.nan])])
        assert primitive_func(array_nans) == (11 / 9.0)

    def test_empty_string(self):
        primitive_func = AverageCountPerUnique().get_function()
        array_empty_string = pd.concat([self.array.copy(), pd.Series([np.nan, "", ""])])
        assert primitive_func(array_empty_string) == (4 / 3.0)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestVariance(PrimitiveTestBase):
    primitive = Variance

    def test_regular(self):
        variance = self.primitive().get_function()
        np.testing.assert_almost_equal(variance(np.array([0, 3, 4, 3])), 2.25)

    def test_single(self):
        variance = self.primitive().get_function()
        np.testing.assert_almost_equal(variance(np.array([4])), 0)

    def test_double(self):
        variance = self.primitive().get_function()
        np.testing.assert_almost_equal(variance(np.array([3, 4])), 0.25)

    def test_empty(self):
        variance = self.primitive().get_function()
        np.testing.assert_almost_equal(variance(np.array([])), np.nan)

    def test_nan(self):
        variance = self.primitive().get_function()
        np.testing.assert_almost_equal(
            variance(pd.Series([0, np.nan, 4, 3])),
            2.8888888888888893,
        )

    def test_allnan(self):
        variance = self.primitive().get_function()
        np.testing.assert_almost_equal(
            variance(pd.Series([np.nan, np.nan, np.nan])),
            np.nan,
        )


class TestFirstLastTimeDelta(PrimitiveTestBase):
    primitive = FirstLastTimeDelta
    times = pd.Series([datetime(2011, 4, 9, 10, 30, i * 6) for i in range(5)])
    actual_delta = (times.iloc[-1] - times.iloc[0]).total_seconds()

    def test_first_last_time_delta(self):
        primitive_func = self.primitive().get_function()
        assert primitive_func(self.times) == self.actual_delta

    def test_with_nans(self):
        primitive_func = self.primitive().get_function()
        times = pd.concat([self.times, pd.Series([np.nan])])
        assert primitive_func(times) == self.actual_delta
        assert pd.isna(primitive_func(pd.Series([np.nan])))

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestEntropy(PrimitiveTestBase):
    primitive = Entropy

    @pytest.mark.parametrize(
        "dtype",
        ["category", "object", "string"],
    )
    def test_regular(self, dtype):
        data = pd.Series([1, 2, 3, 2], dtype=dtype)
        primitive_func = self.primitive().get_function()
        given_answer = primitive_func(data)
        assert np.isclose(given_answer, 1.03, atol=0.01)

    @pytest.mark.parametrize(
        "dtype",
        ["category", "object", "string"],
    )
    def test_empty(self, dtype):
        data = pd.Series([], dtype=dtype)
        primitive_func = self.primitive().get_function()
        given_answer = primitive_func(data)
        assert given_answer == 0.0

    @pytest.mark.parametrize(
        "dtype",
        ["category", "object", "string"],
    )
    def test_args(self, dtype):
        data = pd.Series([1, 2, 3, 2], dtype=dtype)
        if dtype == "string":
            data = pd.concat([data, pd.Series([pd.NA, pd.NA], dtype=dtype)])
        else:
            data = pd.concat([data, pd.Series([np.nan, np.nan], dtype=dtype)])
        primitive_func = self.primitive(dropna=True, base=2).get_function()
        given_answer = primitive_func(data)
        assert np.isclose(given_answer, 1.5, atol=0.001)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive, max_depth=2)


class TestKurtosis(PrimitiveTestBase):
    primitive = Kurtosis

    @pytest.mark.parametrize(
        "dtype",
        ["int64", "float64"],
    )
    def test_regular(self, dtype):
        data = pd.Series([1, 2, 3, 4, 5], dtype=dtype)
        answer = -1.3
        primitive_func = self.primitive().get_function()
        given_answer = primitive_func(data)
        assert np.isclose(answer, given_answer, atol=0.01)

        data = pd.Series([1, 2, 3, 4, 5, 6], dtype=dtype)
        answer = -1.26
        primitive_func = self.primitive().get_function()
        given_answer = primitive_func(data)
        assert np.isclose(answer, given_answer, atol=0.01)

        data = pd.Series([x * x for x in list(range(100))], dtype=dtype)
        answer = -0.85
        primitive_func = self.primitive().get_function()
        given_answer = primitive_func(data)
        assert np.isclose(answer, given_answer, atol=0.01)

        if dtype == "float64":
            # Series contains floating point values - only check with float dtype
            data = pd.Series([sqrt(x) for x in list(range(100))], dtype=dtype)
            answer = -0.46
            primitive_func = self.primitive().get_function()
            given_answer = primitive_func(data)
            assert np.isclose(answer, given_answer, atol=0.01)

    def test_nan(self):
        data = pd.Series([np.nan, 5, 3], dtype="float64")
        primitive_func = self.primitive().get_function()
        given_answer = primitive_func(data)
        assert pd.isna(given_answer)

    @pytest.mark.parametrize(
        "dtype",
        ["int64", "float64"],
    )
    def test_empty(self, dtype):
        data = pd.Series([], dtype=dtype)
        primitive_func = self.primitive().get_function()
        given_answer = primitive_func(data)
        assert pd.isna(given_answer)

    def test_inf(self):
        data = pd.Series([1, np.inf], dtype="float64")
        primitive_func = self.primitive().get_function()
        given_answer = primitive_func(data)
        assert pd.isna(given_answer)

        data = pd.Series([np.NINF, 1, np.inf], dtype="float64")
        primitive_func = self.primitive().get_function()
        given_answer = primitive_func(data)
        assert pd.isna(given_answer)

    def test_arg(self):
        data = pd.Series([1, 2, 3, 4, 5, np.nan, np.nan], dtype="float64")
        answer = -1.3
        primitive_func = self.primitive(nan_policy="omit").get_function()
        given_answer = primitive_func(data)
        assert answer == given_answer

        primitive_func = self.primitive(nan_policy="propagate").get_function()
        given_answer = primitive_func(data)
        assert np.isnan(given_answer)

        primitive_func = self.primitive(nan_policy="raise").get_function()
        with raises(ValueError):
            primitive_func(data)

    def test_error(self):
        with raises(ValueError):
            self.primitive(nan_policy="invalid_policy").get_function()

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestNumZeroCrossings(PrimitiveTestBase):
    primitive = NumZeroCrossings

    def test_nan(self):
        data = pd.Series([3, np.nan, 5, 3, np.nan, 0, np.nan, 0, np.nan, -2])
        # crossing from 0 to np.nan to -2, which is 1 crossing
        answer = 1
        primtive_func = self.primitive().get_function()
        given_answer = primtive_func(data)
        assert given_answer == answer

    def test_empty(self):
        data = pd.Series([], dtype="int64")
        answer = 0
        primtive_func = self.primitive().get_function()
        given_answer = primtive_func(data)
        assert given_answer == answer

    def test_inf(self):
        data = pd.Series([-1, np.inf])
        answer = 1
        primtive_func = self.primitive().get_function()
        given_answer = primtive_func(data)
        assert given_answer == answer

        data = pd.Series([np.NINF, 1, np.inf])
        answer = 1
        primtive_func = self.primitive().get_function()
        given_answer = primtive_func(data)
        assert given_answer == answer

    def test_zeros(self):
        data = pd.Series([1, 0, -1, 0, 1, 0, -1])
        answer = 3
        primtive_func = self.primitive().get_function()
        given_answer = primtive_func(data)
        assert given_answer == answer

        data = pd.Series([1, 0, 1, 0, 1])
        answer = 0
        primtive_func = self.primitive().get_function()
        given_answer = primtive_func(data)
        assert given_answer == answer

    def test_regular(self):
        data = pd.Series([1, 2, 3, 4, 5])
        answer = 0
        primtive_func = self.primitive().get_function()
        given_answer = primtive_func(data)
        assert given_answer == answer

        data = pd.Series([1, -1, 2, -2, 3, -3])
        answer = 5
        primtive_func = self.primitive().get_function()
        given_answer = primtive_func(data)
        assert given_answer == answer

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestNumTrueSinceLastFalse(PrimitiveTestBase):
    primitive = NumTrueSinceLastFalse

    def test_regular(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([False, True, False, True, True])
        answer = primitive_func(bools)
        correct_answer = 2
        assert answer == correct_answer

    def test_regular_end_in_false(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([False, True, False, True, True, False])
        answer = primitive_func(bools)
        correct_answer = 0
        assert answer == correct_answer

    def test_no_false(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([True] * 5)
        assert pd.isna(primitive_func(bools))

    def test_all_false(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([False, False, False])
        answer = primitive_func(bools)
        correct_answer = 0
        assert answer == correct_answer

    def test_nan(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([False, True, np.nan, True, True])
        answer = primitive_func(bools)
        correct_answer = 3
        assert answer == correct_answer

    def test_all_nan(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([np.nan, np.nan, np.nan])
        assert pd.isna(primitive_func(bools))

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestNumFalseSinceLastTrue(PrimitiveTestBase):
    primitive = NumFalseSinceLastTrue

    def test_regular(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([True, False, True, False, False])
        answer = primitive_func(bools)
        correct_answer = 2
        assert answer == correct_answer

    def test_regular_end_in_true(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([True, False, True, False, False, True])
        answer = primitive_func(bools)
        correct_answer = 0
        assert answer == correct_answer

    def test_no_true(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([False] * 5)
        assert pd.isna(primitive_func(bools))

    def test_all_true(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([True, True, True])
        answer = primitive_func(bools)
        correct_answer = 0
        assert answer == correct_answer

    def test_nan(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([True, False, np.nan, False, False])
        answer = primitive_func(bools)
        correct_answer = 3
        assert answer == correct_answer

    def test_all_nan(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([np.nan, np.nan, np.nan])
        assert pd.isna(primitive_func(bools))

    def test_numeric_and_string_input(self):
        primitive_func = self.primitive().get_function()
        bools = pd.Series([True, 0, 1, "10", ""])
        answer = primitive_func(bools)
        correct_answer = 1
        assert answer == correct_answer

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestNumPeaks(PrimitiveTestBase):
    primitive = NumPeaks

    @pytest.mark.parametrize(
        "dtype",
        ["int64", "float64", "Int64"],
    )
    def test_negative_and_positive_nums(self, dtype):
        get_peaks = self.primitive().get_function()
        assert (
            get_peaks(pd.Series([-5, 0, 10, 0, 10, -5, -4, -5, 10, 0], dtype=dtype))
            == 4
        )

    @pytest.mark.parametrize(
        "dtype",
        ["int64", "float64", "Int64"],
    )
    def test_plateu(self, dtype):
        get_peaks = self.primitive().get_function()
        assert get_peaks(pd.Series([1, 2, 3, 3, 3, 3, 3, 2, 1], dtype=dtype)) == 1
        assert get_peaks(pd.Series([1, 2, 3, 3, 3, 4, 3, 3, 3, 2, 1], dtype=dtype)) == 1
        assert (
            get_peaks(
                pd.Series(
                    [
                        5,
                        4,
                        3,
                        3,
                        3,
                        3,
                        3,
                        3,
                        4,
                        5,
                        5,
                        5,
                        5,
                        5,
                        3,
                        3,
                        3,
                        3,
                        4,
                    ],
                    dtype=dtype,
                ),
            )
            == 1
        )
        assert (
            get_peaks(
                pd.Series(
                    [
                        1,
                        2,
                        3,
                        3,
                        3,
                        2,
                        1,
                        2,
                        3,
                        3,
                        3,
                        2,
                        5,
                        5,
                        5,
                        2,
                    ],
                    dtype=dtype,
                ),
            )
            == 3
        )

    @pytest.mark.parametrize(
        "dtype",
        ["int64", "float64", "Int64"],
    )
    def test_regular(self, dtype):
        get_peaks = self.primitive().get_function()
        assert get_peaks(pd.Series([1, 7, 3, 8, 2, 3, 4, 3, 4, 2, 4], dtype=dtype)) == 4
        assert get_peaks(pd.Series([1, 2, 3, 2, 1], dtype=dtype)) == 1

    @pytest.mark.parametrize(
        "dtype",
        ["int64", "float64", "Int64"],
    )
    def test_no_peak(self, dtype):
        get_peaks = self.primitive().get_function()
        assert get_peaks(pd.Series([1, 2, 3], dtype=dtype)) == 0
        assert get_peaks(pd.Series([3, 2, 2, 2, 2, 1], dtype=dtype)) == 0

    @pytest.mark.parametrize(
        "dtype",
        ["int64", "float64", "Int64"],
    )
    def test_too_small_data(self, dtype):
        get_peaks = self.primitive().get_function()
        assert get_peaks(pd.Series([], dtype=dtype)) == 0
        assert get_peaks(pd.Series([1])) == 0
        assert get_peaks(pd.Series([1, 1])) == 0
        assert get_peaks(pd.Series([1, 2])) == 0
        assert get_peaks(pd.Series([2, 1])) == 0

    @pytest.mark.parametrize(
        "dtype",
        ["int64", "float64", "Int64"],
    )
    def test_nans(self, dtype):
        get_peaks = self.primitive().get_function()
        array = pd.Series(
            [
                0,
                5,
                10,
                15,
                20,
                0,
                1,
                2,
                3,
                0,
                0,
                5,
                0,
                7,
                14,
            ],
            dtype=dtype,
        )
        if dtype == "float64":
            array = pd.concat([array, pd.Series([np.nan, np.nan])])
        elif dtype == "Int64":
            array = pd.concat([array, pd.Series([pd.NA, pd.NA])])
        array = array.astype(dtype=dtype)
        assert get_peaks(array) == 3

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestDateFirstEvent(PrimitiveTestBase):
    primitive = DateFirstEvent

    def test_regular(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series(
            [
                "2011-04-09 10:30:00",
                "2011-04-09 10:30:06",
                "2011-04-09 10:30:12",
                "2011-04-09 10:30:18",
            ],
            dtype="datetime64[ns]",
        )
        answer = pd.Timestamp("2011-04-09 10:30:00")
        given_answer = primitive_func(case)
        assert given_answer == answer

    def test_nat(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series(
            [
                pd.NaT,
                pd.NaT,
                "2011-04-09 10:30:12",
                "2011-04-09 10:30:18",
            ],
            dtype="datetime64[ns]",
        )
        answer = pd.Timestamp("2011-04-09 10:30:12")
        given_answer = primitive_func(case)
        assert given_answer == answer

    def test_empty(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([], dtype="datetime64[ns]")
        given_answer = primitive_func(case)
        assert pd.isna(given_answer)

    def test_with_featuretools(self, pd_es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(pd_es, aggregation, transform, self.primitive)

    def test_serialize(self, es):
        check_serialize(self.primitive, es, target_dataframe_name="sessions")


class TestMinCount(PrimitiveTestBase):
    primitive = MinCount

    def test_nan(self):
        data = pd.Series([np.nan, np.nan, np.nan])
        primitive_func = self.primitive().get_function()
        answer = primitive_func(data)
        assert pd.isna(answer)

    def test_inf(self):
        data = pd.Series([5, 10, 10, np.inf, np.inf, np.inf])
        primitive_func = self.primitive().get_function()
        answer = primitive_func(data)
        assert answer == 1

    def test_regular(self):
        data = pd.Series([1, 2, 2, 2, 3, 4, 4, 4, 5])
        primitive_func = self.primitive().get_function()
        answer = primitive_func(data)
        assert answer == 1

        data = pd.Series([2, 2, 2, 3, 4, 4, 4])
        primitive_func = self.primitive().get_function()
        answer = primitive_func(data)
        assert answer == 3

    def test_skipna(self):
        data = pd.Series([1, 1, 2, 3, 4, 4, np.nan, 5])
        primitive_func = self.primitive(skipna=False).get_function()
        answer = primitive_func(data)
        assert pd.isna(answer)

    def test_ninf(self):
        data = pd.Series([np.NINF, np.NINF, np.nan])
        primitive_func = self.primitive().get_function()
        answer = primitive_func(data)
        assert answer == 2

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestMaxCount(PrimitiveTestBase):
    primitive = MaxCount

    def test_nan(self):
        data = pd.Series([np.nan, np.nan, np.nan])
        primitive_func = self.primitive().get_function()
        answer = primitive_func(data)
        assert pd.isna(answer)

    def test_inf(self):
        data = pd.Series([5, 10, 10, np.inf, np.inf, np.inf])
        primitive_func = self.primitive().get_function()
        answer = primitive_func(data)
        assert answer == 3

    def test_regular(self):
        data = pd.Series([1, 1, 2, 3, 4, 4, 4, 5])
        primitive_func = self.primitive().get_function()
        answer = primitive_func(data)
        assert answer == 1

        data = pd.Series([1, 1, 2, 3, 4, 4, 4])
        primitive_func = self.primitive().get_function()
        answer = primitive_func(data)
        assert answer == 3

    def test_skipna(self):
        data = pd.Series([1, 1, 2, 3, 4, 4, np.nan, 5])
        primitive_func = self.primitive(skipna=False).get_function()
        answer = primitive_func(data)
        assert pd.isna(answer)

    def test_ninf(self):
        data = pd.Series([np.NINF, np.NINF, np.nan])
        primitive_func = self.primitive().get_function()
        answer = primitive_func(data)
        assert answer == 2

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestMaxMinDelta(PrimitiveTestBase):
    primitive = MaxMinDelta
    array = pd.Series([1, 1, 2, 2, 3, 4, 5, 6, 7, 8])

    def test_max_min_delta(self):
        primitive_func = self.primitive().get_function()
        assert primitive_func(self.array) == 7.0

    def test_nans(self):
        primitive_func = self.primitive().get_function()
        array_nans = pd.concat([self.array, pd.Series([np.nan])])
        assert primitive_func(array_nans) == 7.0
        primitive_func = self.primitive(skipna=False).get_function()
        array_nans = pd.concat([self.array, pd.Series([np.nan])])
        assert pd.isna(primitive_func(array_nans))

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestMedianCount(PrimitiveTestBase):
    primitive = MedianCount

    def test_regular(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([1, 3, 5, 7])
        given_answer = primitive_func(case)
        assert given_answer == 0

    def test_nans(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([1, 3, 4, 4, 4, 5, 7, np.nan, np.nan])
        given_answer = primitive_func(case)
        assert given_answer == 3
        primitive_func = self.primitive(skipna=False).get_function()
        given_answer = primitive_func(case)
        assert pd.isna(given_answer)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestNMostCommonFrequency(PrimitiveTestBase):
    primitive = NMostCommonFrequency

    def test_regular(self):
        test_cases = [
            pd.Series([8, 7, 10, 10, 10, 3, 4, 5, 10, 8, 7]),
            pd.Series([7, 7, 7, 6, 6, 5, 4]),
            pd.Series([4, 5, 6, 6, 7, 7, 7]),
        ]

        answers = [
            pd.Series([4, 2, 2]),
            pd.Series([3, 2, 1]),
            pd.Series([3, 2, 1]),
        ]

        primtive_func = self.primitive(3).get_function()

        for case, answer in zip(test_cases, answers):
            given_answer = primtive_func(case)
            given_answer = given_answer.reset_index(drop=True)
            assert given_answer.equals(answer)

    def test_n_larger_than_len(self):
        test_cases = [
            pd.Series(["red", "red", "blue", "green"]),
            pd.Series(["red", "red", "red", "blue", "green"]),
            pd.Series(["red", "blue", "green", "orange"]),
        ]
        answers = [
            pd.Series([2, 1, 1, np.nan, np.nan]),
            pd.Series([3, 1, 1, np.nan, np.nan]),
            pd.Series([1, 1, 1, 1, np.nan]),
        ]

        primtive_func = self.primitive(5).get_function()
        for case, answer in zip(test_cases, answers):
            given_answer = primtive_func(case)
            given_answer = given_answer.reset_index(drop=True)
            assert given_answer.equals(answer)

    def test_skipna(self):
        array = pd.Series(["red", "red", "blue", "green", np.nan, np.nan])
        primtive_func = self.primitive(5, skipna=False).get_function()
        given_answer = primtive_func(array)
        given_answer = given_answer.reset_index(drop=True)
        answer = pd.Series([2, 2, 1, 1, np.nan])
        assert given_answer.equals(answer)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        aggregation.append(self.primitive(5))
        valid_dfs(
            es,
            aggregation,
            transform,
            self.primitive,
            target_dataframe_name="customers",
            multi_output=True,
        )

    def test_with_featuretools_args(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        aggregation.append(self.primitive(5, skipna=False))
        valid_dfs(
            es,
            aggregation,
            transform,
            self.primitive,
            target_dataframe_name="customers",
            multi_output=True,
        )

    def test_serialize(self, es):
        check_serialize(
            primitive=self.primitive,
            es=es,
            target_dataframe_name="customers",
        )


class TestNUniqueDays(PrimitiveTestBase):
    primitive = NUniqueDays

    def test_two_years(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2011-12-31"))
        assert primitive_func(array) == 365 * 2

    def test_leap_year(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2016-01-01", "2017-12-31"))
        assert primitive_func(array) == 365 * 2 + 1

    def test_ten_years(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2019-12-31"))
        assert primitive_func(array) == 365 * 10 + 1 + 1

    def test_distinct_dt(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                datetime(2019, 2, 21),
                datetime(2019, 2, 1, 1, 20, 0),
                datetime(2019, 2, 1, 1, 30, 0),
                datetime(2018, 2, 1),
                datetime(2019, 1, 1),
            ],
        )
        assert primitive_func(array) == 4

    def test_NaT(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2011-12-31"))
        NaT_array = pd.Series([pd.NaT] * 100)
        assert primitive_func(pd.concat([array, NaT_array])) == 365 * 2

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestNUniqueDaysOfCalendarYear(PrimitiveTestBase):
    primitive = NUniqueDaysOfCalendarYear

    def test_two_years(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2011-12-31"))
        assert primitive_func(array) == 365

    def test_leap_year(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2016-01-01", "2017-12-31"))
        assert primitive_func(array) == 366

    def test_ten_years(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2019-12-31"))
        assert primitive_func(array) == 366

    def test_distinct_dt(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                datetime(2019, 2, 21),
                datetime(2019, 2, 1, 1, 20, 0),
                datetime(2019, 2, 1, 1, 30, 0),
                datetime(2018, 2, 1),
                datetime(2019, 1, 1),
            ],
        )
        assert primitive_func(array) == 3

    def test_NaT(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2011-12-31"))
        NaT_array = pd.Series([pd.NaT] * 100)
        assert primitive_func(pd.concat([array, NaT_array])) == 365

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestNUniqueDaysOfMonth(PrimitiveTestBase):
    primitive = NUniqueDaysOfMonth

    def test_two_days(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2010-01-02"))
        assert primitive_func(array) == 2

    def test_one_year(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2010-12-31"))
        assert primitive_func(array) == 31

    def test_leap_year(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2016-01-01", "2017-12-31"))
        assert primitive_func(array) == 31

    def test_distinct_dt(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                datetime(2019, 2, 21),
                datetime(2019, 2, 1, 1, 20, 0),
                datetime(2019, 2, 1, 1, 30, 0),
                datetime(2018, 2, 1),
                datetime(2019, 1, 1),
            ],
        )
        assert primitive_func(array) == 2

    def test_NaT(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2010-12-31"))
        NaT_array = pd.Series([pd.NaT] * 100)
        assert primitive_func(pd.concat([array, NaT_array])) == 31

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestNUniqueMonths(PrimitiveTestBase):
    primitive = NUniqueMonths

    def test_two_days(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2010-01-02"))
        assert primitive_func(array) == 1

    def test_ten_years(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2019-12-31"))
        assert primitive_func(array) == 12 * 10

    def test_distinct_dt(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                datetime(2019, 2, 21),
                datetime(2019, 2, 1, 1, 20, 0),
                datetime(2019, 2, 1, 1, 30, 0),
                datetime(2018, 2, 1),
                datetime(2019, 1, 1),
            ],
        )
        assert primitive_func(array) == 3

    def test_NaT(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2011-12-31"))
        NaT_array = pd.Series([pd.NaT] * 100)
        assert primitive_func(pd.concat([array, NaT_array])) == 12 * 2

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestNUniqueWeeks(PrimitiveTestBase):
    primitive = NUniqueWeeks

    def test_same_week(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2019-01-01", "2019-01-02"))
        assert primitive_func(array) == 1

    def test_ten_years(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2010-01-01", "2019-12-31"))
        assert primitive_func(array) == 523

    def test_distinct_dt(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                datetime(2019, 2, 21),
                datetime(2019, 2, 1, 1, 20, 0),
                datetime(2019, 2, 1, 1, 30, 0),
                datetime(2018, 2, 2),
                datetime(2019, 2, 3, 1, 30, 0),
                datetime(2019, 1, 1),
            ],
        )
        assert primitive_func(array) == 4

    def test_NaT(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(pd.date_range("2019-01-01", "2019-01-02"))
        NaT_array = pd.Series([pd.NaT] * 100)
        assert primitive_func(pd.concat([array, NaT_array])) == 1

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        aggregation.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestHasNoDuplicates(PrimitiveTestBase):
    primitive = HasNoDuplicates

    def test_regular(self):
        primitive_func = self.primitive().get_function()
        data = pd.Series([1, 1, 2])
        assert not primitive_func(data)
        assert isinstance(primitive_func(data), bool)

        data = pd.Series([1, 2, 3])
        assert primitive_func(data)
        assert isinstance(primitive_func(data), bool)

        data = pd.Series([1, 2, 4])
        assert primitive_func(data)
        assert isinstance(primitive_func(data), bool)

        data = pd.Series(["red", "blue", "orange"])
        assert primitive_func(data)
        assert isinstance(primitive_func(data), bool)

        data = pd.Series(["red", "blue", "red"])
        assert not primitive_func(data)

    def test_nan(self):
        primitive_func = self.primitive().get_function()
        data = pd.Series([np.nan, 1, 2, 3])
        assert primitive_func(data)
        assert isinstance(primitive_func(data), bool)

        data = pd.Series([np.nan, np.nan, 1])
        # drop both nans, so has 1 value
        assert primitive_func(data) is True
        assert isinstance(primitive_func(data), bool)

        primitive_func = self.primitive(skipna=False).get_function()
        data = pd.Series([np.nan, np.nan, 1])
        assert primitive_func(data) is False
        assert isinstance(primitive_func(data), bool)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instantiate = self.primitive()
        aggregation.append(primitive_instantiate)
        valid_dfs(
            es,
            aggregation,
            transform,
            self.primitive,
            target_dataframe_name="customers",
            instance_ids=[0, 1, 2],
        )


class TestIsMonotonicallyDecreasing(PrimitiveTestBase):
    primitive = IsMonotonicallyDecreasing

    def test_monotonically_decreasing(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([9, 5, 3, 1, -1])
        assert primitive_func(case) is True

    def test_monotonically_increasing(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([-1, 1, 3, 5, 9])
        assert primitive_func(case) is False

    def test_non_monotonic(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([-1, 1, 3, 2, 5])
        assert primitive_func(case) is False

    def test_weakly_decreasing(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([9, 3, 3, 1, -1])
        assert primitive_func(case) is True

    def test_nan(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([9, 5, 3, np.nan, 1, -1])
        assert primitive_func(case) is True

        primitive_func = self.primitive().get_function()
        case = pd.Series([-1, 1, 3, np.nan, 5, 9])
        assert primitive_func(case) is False

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instantiate = self.primitive()
        aggregation.append(primitive_instantiate)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestIsMonotonicallyIncreasing(PrimitiveTestBase):
    primitive = IsMonotonicallyIncreasing

    def test_monotonically_increasing(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([-1, 1, 3, 5, 9])
        assert primitive_func(case) is True

    def test_monotonically_decreasing(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([9, 5, 3, 1, -1])
        assert primitive_func(case) is False

    def test_non_monotonic(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([-1, 1, 3, 2, 5])
        assert primitive_func(case) is False

    def test_weakly_increasing(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([-1, 1, 3, 3, 9])
        assert primitive_func(case) is True

    def test_nan(self):
        primitive_func = self.primitive().get_function()
        case = pd.Series([-1, 1, 3, np.nan, 5, 9])
        assert primitive_func(case) is True

        primitive_func = self.primitive().get_function()
        case = pd.Series([9, 5, 3, np.nan, 1, -1])
        assert primitive_func(case) is False

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instantiate = self.primitive()
        aggregation.append(primitive_instantiate)
        valid_dfs(es, aggregation, transform, self.primitive)



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Boolean, BooleanNullable

from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class MultiplyBoolean(TransformPrimitive):
    """Performs element-wise multiplication of two lists of boolean values.

    Description:
        Given a list of boolean values X and a list of boolean
        values Y, determine the product of each value in X
        with its corresponding value in Y.

    Examples:
        >>> multiply_boolean = MultiplyBoolean()
        >>> multiply_boolean([True, True, False], [True, False, True]).tolist()
        [True, False, False]
    """

    name = "multiply_boolean"
    input_types = [
        [
            ColumnSchema(logical_type=BooleanNullable),
            ColumnSchema(logical_type=BooleanNullable),
        ],
        [ColumnSchema(logical_type=Boolean), ColumnSchema(logical_type=Boolean)],
        [
            ColumnSchema(logical_type=Boolean),
            ColumnSchema(logical_type=BooleanNullable),
        ],
        [
            ColumnSchema(logical_type=BooleanNullable),
            ColumnSchema(logical_type=Boolean),
        ],
    ]
    return_type = ColumnSchema(logical_type=BooleanNullable)
    commutative = True
    compatibility = [Library.PANDAS, Library.DASK]
    description_template = "the product of {} and {}"

    def get_function(self):
        return np.bitwise_and

    def generate_name(self, base_feature_names):
        return "%s * %s" % (base_feature_names[0], base_feature_names[1])



from datetime import datetime

import numpy as np
import pandas as pd
import pytest
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Categorical, Datetime, Double, Integer

from featuretools import (
    AggregationFeature,
    EntitySet,
    Feature,
    Timedelta,
    calculate_feature_matrix,
)
from featuretools.computational_backends.feature_set import FeatureSet
from featuretools.computational_backends.feature_set_calculator import (
    FeatureSetCalculator,
)
from featuretools.entityset.relationship import RelationshipPath
from featuretools.feature_base import DirectFeature, IdentityFeature
from featuretools.primitives import (
    And,
    Count,
    CumSum,
    EqualScalar,
    GreaterThanEqualToScalar,
    GreaterThanScalar,
    LessThanEqualToScalar,
    LessThanScalar,
    Mean,
    Min,
    Mode,
    Negate,
    NMostCommon,
    NotEqualScalar,
    NumTrue,
    Sum,
    TimeSinceLast,
    Trend,
)
from featuretools.primitives.base import AggregationPrimitive
from featuretools.tests.testing_utils import backward_path, to_pandas
from featuretools.utils import Trie
from featuretools.utils.gen_utils import Library, import_or_none, is_instance

dd = import_or_none("dask.dataframe")


def test_make_identity(es):
    f = IdentityFeature(es["log"].ww["datetime"])

    feature_set = FeatureSet([f])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = to_pandas(calculator.run(np.array([0])))

    v = df[f.get_name()][0]
    assert v == datetime(2011, 4, 9, 10, 30, 0)


def test_make_dfeat(es):
    f = DirectFeature(
        Feature(es["customers"].ww["age"]),
        child_dataframe_name="sessions",
    )

    feature_set = FeatureSet([f])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = to_pandas(calculator.run(np.array([0])))

    v = df[f.get_name()][0]
    assert v == 33


def test_make_agg_feat_of_identity_column(es):
    agg_feat = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="sessions",
        primitive=Sum,
    )

    feature_set = FeatureSet([agg_feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = to_pandas(calculator.run(np.array([0])))

    v = df[agg_feat.get_name()][0]
    assert v == 50


# full_dataframe not supported with Dask
def test_full_dataframe_trans_of_agg(pd_es):
    agg_feat = Feature(
        pd_es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Sum,
    )
    trans_feat = Feature(agg_feat, primitive=CumSum)

    feature_set = FeatureSet([trans_feat])
    calculator = FeatureSetCalculator(pd_es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([1]))

    v = df[trans_feat.get_name()].values[0]
    assert v == 82


def test_full_dataframe_error_dask(dask_es):
    agg_feat = Feature(
        dask_es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Sum,
    )
    trans_feat = Feature(agg_feat, primitive=CumSum)

    feature_set = FeatureSet([trans_feat])
    calculator = FeatureSetCalculator(dask_es, time_last=None, feature_set=feature_set)
    error_text = "Cannot use primitives that require full dataframe with Dask"

    with pytest.raises(ValueError, match=error_text):
        calculator.run(np.array([1]))


def test_make_agg_feat_of_identity_index_column(es):
    agg_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        primitive=Count,
    )

    feature_set = FeatureSet([agg_feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = to_pandas(calculator.run(np.array([0])))

    v = df[agg_feat.get_name()][0]
    assert v == 5


def test_make_agg_feat_where_count(es):
    agg_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        where=IdentityFeature(es["log"].ww["product_id"]) == "coke zero",
        primitive=Count,
    )

    feature_set = FeatureSet([agg_feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = to_pandas(calculator.run(np.array([0])))

    v = df[agg_feat.get_name()][0]
    assert v == 3


def test_make_agg_feat_using_prev_time(es):
    agg_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        use_previous=Timedelta(10, "s"),
        primitive=Count,
    )

    feature_set = FeatureSet([agg_feat])
    calculator = FeatureSetCalculator(
        es,
        time_last=datetime(2011, 4, 9, 10, 30, 10),
        feature_set=feature_set,
    )
    df = to_pandas(calculator.run(np.array([0])))

    v = df[agg_feat.get_name()][0]
    assert v == 2

    calculator = FeatureSetCalculator(
        es,
        time_last=datetime(2011, 4, 9, 10, 30, 30),
        feature_set=feature_set,
    )
    df = to_pandas(calculator.run(np.array([0])))

    v = df[agg_feat.get_name()][0]
    assert v == 1


def test_make_agg_feat_using_prev_n_events(es):
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail("Distrubuted entitysets do not support use_previous")
    agg_feat_1 = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="sessions",
        use_previous=Timedelta(1, "observations"),
        primitive=Min,
    )

    agg_feat_2 = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="sessions",
        use_previous=Timedelta(3, "observations"),
        primitive=Min,
    )

    assert (
        agg_feat_1.get_name() != agg_feat_2.get_name()
    ), "Features should have different names based on use_previous"

    feature_set = FeatureSet([agg_feat_1, agg_feat_2])
    calculator = FeatureSetCalculator(
        es,
        time_last=datetime(2011, 4, 9, 10, 30, 6),
        feature_set=feature_set,
    )
    df = calculator.run(np.array([0]))

    # time_last is included by default
    v1 = df[agg_feat_1.get_name()][0]
    v2 = df[agg_feat_2.get_name()][0]
    assert v1 == 5
    assert v2 == 0

    calculator = FeatureSetCalculator(
        es,
        time_last=datetime(2011, 4, 9, 10, 30, 30),
        feature_set=feature_set,
    )
    df = calculator.run(np.array([0]))

    v1 = df[agg_feat_1.get_name()][0]
    v2 = df[agg_feat_2.get_name()][0]
    assert v1 == 20
    assert v2 == 10


def test_make_agg_feat_multiple_dtypes(es):
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail(
            "Currently no Dask or Spark compatible agg prims that use multiple dtypes",
        )
    compare_prod = IdentityFeature(es["log"].ww["product_id"]) == "coke zero"

    agg_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        where=compare_prod,
        primitive=Count,
    )

    agg_feat2 = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="sessions",
        where=compare_prod,
        primitive=Mode,
    )

    feature_set = FeatureSet([agg_feat, agg_feat2])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0]))

    v = df[agg_feat.get_name()][0]
    v2 = df[agg_feat2.get_name()][0]
    assert v == 3
    assert v2 == "coke zero"


def test_make_agg_feat_where_different_identity_feat(es):
    feats = []
    where_cmps = [
        LessThanScalar,
        GreaterThanScalar,
        LessThanEqualToScalar,
        GreaterThanEqualToScalar,
        EqualScalar,
        NotEqualScalar,
    ]
    for where_cmp in where_cmps:
        feats.append(
            Feature(
                es["log"].ww["id"],
                parent_dataframe_name="sessions",
                where=Feature(
                    es["log"].ww["value"],
                    primitive=where_cmp(10.0),
                ),
                primitive=Count,
            ),
        )

    df = calculate_feature_matrix(
        entityset=es,
        features=feats,
        instance_ids=[0, 1, 2, 3],
    )
    df = to_pandas(df, index="id", sort_index=True)

    for i, where_cmp in enumerate(where_cmps):
        name = feats[i].get_name()
        instances = df[name]
        v0, v1, v2, v3 = instances[0:4]
        if where_cmp == LessThanScalar:
            assert v0 == 2
            assert v1 == 4
            assert v2 == 1
            assert v3 == 2
        elif where_cmp == GreaterThanScalar:
            assert v0 == 2
            assert v1 == 0
            assert v2 == 0
            assert v3 == 0
        elif where_cmp == LessThanEqualToScalar:
            assert v0 == 3
            assert v1 == 4
            assert v2 == 1
            assert v3 == 2
        elif where_cmp == GreaterThanEqualToScalar:
            assert v0 == 3
            assert v1 == 0
            assert v2 == 0
            assert v3 == 0
        elif where_cmp == EqualScalar:
            assert v0 == 1
            assert v1 == 0
            assert v2 == 0
            assert v3 == 0
        elif where_cmp == NotEqualScalar:
            assert v0 == 4
            assert v1 == 4
            assert v2 == 1
            assert v3 == 2


def test_make_agg_feat_of_grandchild_dataframe(es):
    agg_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="customers",
        primitive=Count,
    )

    feature_set = FeatureSet([agg_feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0]))
    df = to_pandas(df, index="id")
    v = df[agg_feat.get_name()].values[0]
    assert v == 10


def test_make_agg_feat_where_count_feat(es):
    """
    Feature we're creating is:
    Number of sessions for each customer where the
    number of logs in the session is less than 3
    """
    log_count_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        primitive=Count,
    )

    feat = Feature(
        es["sessions"].ww["id"],
        parent_dataframe_name="customers",
        where=log_count_feat > 1,
        primitive=Count,
    )

    feature_set = FeatureSet([feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0, 1]))
    df = to_pandas(df, index="id", sort_index=True)

    name = feat.get_name()
    instances = df[name]
    v0, v1 = instances[0:2]
    assert v0 == 2
    assert v1 == 2


def test_make_compare_feat(es):
    """
    Feature we're creating is:
    Number of sessions for each customer where the
    number of logs in the session is less than 3
    """
    log_count_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        primitive=Count,
    )

    mean_agg_feat = Feature(
        log_count_feat,
        parent_dataframe_name="customers",
        primitive=Mean,
    )

    mean_feat = DirectFeature(mean_agg_feat, child_dataframe_name="sessions")

    feat = log_count_feat > mean_feat

    feature_set = FeatureSet([feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0, 1, 2]))
    df = to_pandas(df, index="id", sort_index=True)

    name = feat.get_name()
    instances = df[name]
    v0, v1, v2 = instances[0:3]
    assert v0
    assert v1
    assert not v2


def test_make_agg_feat_where_count_and_device_type_feat(es):
    """
    Feature we're creating is:
    Number of sessions for each customer where the
    number of logs in the session is less than 3
    """
    log_count_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        primitive=Count,
    )

    compare_count = log_count_feat == 1
    compare_device_type = IdentityFeature(es["sessions"].ww["device_type"]) == 1
    and_feat = Feature([compare_count, compare_device_type], primitive=And)
    feat = Feature(
        es["sessions"].ww["id"],
        parent_dataframe_name="customers",
        where=and_feat,
        primitive=Count,
    )

    feature_set = FeatureSet([feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0]))
    df = to_pandas(df, index="id")

    name = feat.get_name()
    instances = df[name]
    assert instances.values[0] == 1


def test_make_agg_feat_where_count_or_device_type_feat(es):
    """
    Feature we're creating is:
    Number of sessions for each customer where the
    number of logs in the session is less than 3
    """
    log_count_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        primitive=Count,
    )

    compare_count = log_count_feat > 1
    compare_device_type = IdentityFeature(es["sessions"].ww["device_type"]) == 1
    or_feat = compare_count.OR(compare_device_type)
    feat = Feature(
        es["sessions"].ww["id"],
        parent_dataframe_name="customers",
        where=or_feat,
        primitive=Count,
    )

    feature_set = FeatureSet([feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0]))
    df = to_pandas(df, index="id", int_index=True)

    name = feat.get_name()
    instances = df[name]
    assert instances.values[0] == 3


def test_make_agg_feat_of_agg_feat(es):
    log_count_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        primitive=Count,
    )

    customer_sum_feat = Feature(
        log_count_feat,
        parent_dataframe_name="customers",
        primitive=Sum,
    )

    feature_set = FeatureSet([customer_sum_feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0]))
    df = to_pandas(df, index="id")
    v = df[customer_sum_feat.get_name()].values[0]
    assert v == 10


@pytest.fixture
def pd_df():
    return pd.DataFrame(
        {
            "id": ["a", "b", "c", "d", "e"],
            "e1": ["h", "h", "i", "i", "j"],
            "e2": ["x", "x", "y", "y", "x"],
            "e3": ["z", "z", "z", "z", "z"],
            "val": [1, 1, 1, 1, 1],
        },
    )


@pytest.fixture
def dd_df(pd_df):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(pd_df, npartitions=2)


@pytest.fixture
def spark_df(pd_df):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(pd_df)


@pytest.fixture(params=["pd_df", "dd_df", "spark_df"])
def df(request):
    return request.getfixturevalue(request.param)


def test_make_3_stacked_agg_feats(df):
    """
    Tests stacking 3 agg features.

    The test specifically uses non numeric indices to test how ancestor columns are handled
    as dataframes are merged together

    """
    if is_instance(df, dd, "DataFrame"):
        pytest.xfail("normalize_datdataframe fails with dask DataFrame")
    es = EntitySet()
    ltypes = {"e1": Categorical, "e2": Categorical, "e3": Categorical, "val": Double}
    es.add_dataframe(
        dataframe=df,
        index="id",
        dataframe_name="e0",
        logical_types=ltypes,
    )

    es.normalize_dataframe(
        base_dataframe_name="e0",
        new_dataframe_name="e1",
        index="e1",
        additional_columns=["e2", "e3"],
    )

    es.normalize_dataframe(
        base_dataframe_name="e1",
        new_dataframe_name="e2",
        index="e2",
        additional_columns=["e3"],
    )

    es.normalize_dataframe(
        base_dataframe_name="e2",
        new_dataframe_name="e3",
        index="e3",
    )

    sum_1 = Feature(es["e0"].ww["val"], parent_dataframe_name="e1", primitive=Sum)
    sum_2 = Feature(sum_1, parent_dataframe_name="e2", primitive=Sum)
    sum_3 = Feature(sum_2, parent_dataframe_name="e3", primitive=Sum)

    feature_set = FeatureSet([sum_3])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array(["z"]))
    v = df[sum_3.get_name()][0]
    assert v == 5


def test_make_dfeat_of_agg_feat_on_self(es):
    """
    The graph looks like this:

        R       R = Regions, a parent of customers
        |
        C       C = Customers, the dataframe we're trying to predict on
        |
       etc.

    We're trying to calculate a DFeat from C to R on an agg_feat of R on C.
    """
    customer_count_feat = Feature(
        es["customers"].ww["id"],
        parent_dataframe_name="régions",
        primitive=Count,
    )

    num_customers_feat = DirectFeature(
        customer_count_feat,
        child_dataframe_name="customers",
    )

    feature_set = FeatureSet([num_customers_feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0]))
    df = to_pandas(df, index="id")
    v = df[num_customers_feat.get_name()].values[0]
    assert v == 3


def test_make_dfeat_of_agg_feat_through_parent(es):
    """
    The graph looks like this:

        R       C = Customers, the dataframe we're trying to predict on
       / \\     R = Regions, a parent of customers
      S   C     S = Stores, a child of regions
          |
         etc.

    We're trying to calculate a DFeat from C to R on an agg_feat of R on S.
    """
    store_id_feat = IdentityFeature(es["stores"].ww["id"])

    store_count_feat = Feature(
        store_id_feat,
        parent_dataframe_name="régions",
        primitive=Count,
    )

    num_stores_feat = DirectFeature(store_count_feat, child_dataframe_name="customers")

    feature_set = FeatureSet([num_stores_feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0]))
    df = to_pandas(df, index="id")
    v = df[num_stores_feat.get_name()].values[0]
    assert v == 3


def test_make_deep_agg_feat_of_dfeat_of_agg_feat(es):
    """
    The graph looks like this (higher implies parent):

          C     C = Customers, the dataframe we're trying to predict on
          |     S = Sessions, a child of Customers
      P   S     L = Log, a child of both Sessions and Log
       \\ /     P = Products, a parent of Log which is not a descendent of customers
        L

    We're trying to calculate a DFeat from L to P on an agg_feat of P on L, and
    then aggregate it with another agg_feat of C on L.
    """
    log_count_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="products",
        primitive=Count,
    )

    product_purchases_feat = DirectFeature(log_count_feat, child_dataframe_name="log")

    purchase_popularity = Feature(
        product_purchases_feat,
        parent_dataframe_name="customers",
        primitive=Mean,
    )

    feature_set = FeatureSet([purchase_popularity])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0]))
    df = to_pandas(df, index="id")
    v = df[purchase_popularity.get_name()].values[0]
    assert v == 38.0 / 10.0


def test_deep_agg_feat_chain(es):
    """
    Agg feat of agg feat:
        region.Mean(customer.Count(Log))
    """
    customer_count_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="customers",
        primitive=Count,
    )

    region_avg_feat = Feature(
        customer_count_feat,
        parent_dataframe_name="régions",
        primitive=Mean,
    )

    feature_set = FeatureSet([region_avg_feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array(["United States"]))
    df = to_pandas(df, index="id")

    v = df[region_avg_feat.get_name()][0]
    assert v == 17 / 3.0


# NMostCommon not supported with Dask or Spark
def test_topn(pd_es):
    topn = Feature(
        pd_es["log"].ww["product_id"],
        parent_dataframe_name="customers",
        primitive=NMostCommon(n=2),
    )
    feature_set = FeatureSet([topn])

    calculator = FeatureSetCalculator(pd_es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0, 1, 2]))
    true_results = pd.DataFrame(
        [
            ["toothpaste", "coke zero"],
            ["coke zero", "Haribo sugar-free gummy bears"],
            ["taco clock", np.nan],
        ],
    )
    assert [name in df.columns for name in topn.get_feature_names()]

    for i in range(df.shape[0]):
        true = true_results.loc[i]
        actual = df.loc[i]
        if i == 0:
            # coke zero and toothpase have same number of occurrences
            assert set(true.values) == set(actual.values)
        else:
            for i1, i2 in zip(true, actual):
                assert (pd.isnull(i1) and pd.isnull(i2)) or (i1 == i2)


# Trend not supported with Dask or Spark
def test_trend(pd_es):
    trend = Feature(
        [Feature(pd_es["log"].ww["value"]), Feature(pd_es["log"].ww["datetime"])],
        parent_dataframe_name="customers",
        primitive=Trend,
    )
    feature_set = FeatureSet([trend])

    calculator = FeatureSetCalculator(pd_es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0, 1, 2]))

    true_results = [-0.812730, 4.870378, np.nan]

    np.testing.assert_almost_equal(
        df[trend.get_name()].tolist(),
        true_results,
        decimal=5,
    )


def test_direct_squared(es):
    feature = IdentityFeature(es["log"].ww["value"])
    squared = feature * feature
    feature_set = FeatureSet([feature, squared])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = to_pandas(calculator.run(np.array([0, 1, 2])))
    for i, row in df.iterrows():
        assert (row[0] * row[0]) == row[1]


def test_agg_empty_child(es):
    customer_count_feat = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="customers",
        primitive=Count,
    )
    feature_set = FeatureSet([customer_count_feat])

    # time last before the customer had any events, so child frame is empty
    calculator = FeatureSetCalculator(
        es,
        time_last=datetime(2011, 4, 8),
        feature_set=feature_set,
    )
    df = to_pandas(calculator.run(np.array([0])), index="id")

    assert df["COUNT(log)"].iloc[0] == 0


def test_diamond_entityset(diamond_es):
    es = diamond_es

    amount = IdentityFeature(es["transactions"].ww["amount"])
    path = backward_path(es, ["regions", "customers", "transactions"])
    through_customers = AggregationFeature(
        amount,
        "regions",
        primitive=Sum,
        relationship_path=path,
    )
    path = backward_path(es, ["regions", "stores", "transactions"])
    through_stores = AggregationFeature(
        amount,
        "regions",
        primitive=Sum,
        relationship_path=path,
    )

    feature_set = FeatureSet([through_customers, through_stores])
    calculator = FeatureSetCalculator(
        es,
        time_last=datetime(2011, 4, 8),
        feature_set=feature_set,
    )
    df = calculator.run(np.array([0, 1, 2]))
    df = to_pandas(df, index="id", sort_index=True)

    assert (df["SUM(stores.transactions.amount)"] == [94, 261, 128]).all()
    assert (df["SUM(customers.transactions.amount)"] == [72, 411, 0]).all()


def test_two_relationships_to_single_dataframe(games_es):
    es = games_es
    home_team, away_team = es.relationships
    path = RelationshipPath([(False, home_team)])
    mean_at_home = AggregationFeature(
        Feature(es["games"].ww["home_team_score"]),
        "teams",
        relationship_path=path,
        primitive=Mean,
    )
    path = RelationshipPath([(False, away_team)])
    mean_at_away = AggregationFeature(
        Feature(es["games"].ww["away_team_score"]),
        "teams",
        relationship_path=path,
        primitive=Mean,
    )
    home_team_mean = DirectFeature(mean_at_home, "games", relationship=home_team)
    away_team_mean = DirectFeature(mean_at_away, "games", relationship=away_team)

    feature_set = FeatureSet([home_team_mean, away_team_mean])
    calculator = FeatureSetCalculator(
        es,
        time_last=datetime(2011, 8, 28),
        feature_set=feature_set,
    )
    df = calculator.run(np.array(range(3)))
    df = to_pandas(df, index="id", sort_index=True)

    assert (df[home_team_mean.get_name()] == [1.5, 1.5, 2.5]).all()
    assert (df[away_team_mean.get_name()] == [1, 0.5, 2]).all()


@pytest.fixture
def pd_parent_child():
    parent_df = pd.DataFrame({"id": [1]})
    child_df = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "parent_id": [1, 1, 1],
            "time_index": pd.date_range(start="1/1/2018", periods=3),
            "value": [10, 5, 2],
            "cat": ["a", "a", "b"],
        },
    ).astype({"cat": "category"})
    return (parent_df, child_df)


@pytest.fixture
def dd_parent_child(pd_parent_child):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    parent_df, child_df = pd_parent_child
    parent_df = dd.from_pandas(parent_df, npartitions=2)
    child_df = dd.from_pandas(child_df, npartitions=2)
    return (parent_df, child_df)


@pytest.fixture
def spark_parent_child(pd_parent_child):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    parent_df, child_df = pd_parent_child
    parent_df = ps.from_pandas(parent_df)
    child_df = ps.from_pandas(child_df)
    return (parent_df, child_df)


@pytest.fixture(params=["pd_parent_child", "dd_parent_child", "spark_parent_child"])
def parent_child(request):
    return request.getfixturevalue(request.param)


def test_empty_child_dataframe(parent_child):
    parent_df, child_df = parent_child
    child_ltypes = {
        "parent_id": Integer,
        "time_index": Datetime,
        "value": Double,
        "cat": Categorical,
    }

    es = EntitySet(id="blah")
    es.add_dataframe(dataframe_name="parent", dataframe=parent_df, index="id")
    es.add_dataframe(
        dataframe_name="child",
        dataframe=child_df,
        index="id",
        time_index="time_index",
        logical_types=child_ltypes,
    )
    es.add_relationship("parent", "id", "child", "parent_id")

    # create regular agg
    count = Feature(
        es["child"].ww["id"],
        parent_dataframe_name="parent",
        primitive=Count,
    )

    # create agg feature that requires multiple arguments
    trend = Feature(
        [Feature(es["child"].ww["value"]), Feature(es["child"].ww["time_index"])],
        parent_dataframe_name="parent",
        primitive=Trend,
    )

    # create multi-output agg feature
    n_most_common = Feature(
        es["child"].ww["cat"],
        parent_dataframe_name="parent",
        primitive=NMostCommon,
    )

    # create aggs with where
    where = Feature(es["child"].ww["value"]) == 1
    count_where = Feature(
        es["child"].ww["id"],
        parent_dataframe_name="parent",
        where=where,
        primitive=Count,
    )
    trend_where = Feature(
        [Feature(es["child"].ww["value"]), Feature(es["child"].ww["time_index"])],
        parent_dataframe_name="parent",
        where=where,
        primitive=Trend,
    )
    n_most_common_where = Feature(
        es["child"].ww["cat"],
        parent_dataframe_name="parent",
        where=where,
        primitive=NMostCommon,
    )

    if isinstance(parent_df, pd.DataFrame):
        features = [
            count,
            count_where,
            trend,
            trend_where,
            n_most_common,
            n_most_common_where,
        ]
        data = {
            count.get_name(): pd.Series([0], dtype="Int64"),
            count_where.get_name(): pd.Series([0], dtype="Int64"),
            trend.get_name(): pd.Series([np.nan], dtype="float"),
            trend_where.get_name(): pd.Series([np.nan], dtype="float"),
        }
        for name in n_most_common.get_feature_names():
            data[name] = pd.Series([np.nan], dtype="category")
        for name in n_most_common_where.get_feature_names():
            data[name] = pd.Series([np.nan], dtype="category")
    else:
        features = [count, count_where]
        data = {
            count.get_name(): pd.Series([0], dtype="Int64"),
            count_where.get_name(): pd.Series([0], dtype="Int64"),
        }

    answer = pd.DataFrame(data)

    # cutoff time before all rows
    fm = calculate_feature_matrix(
        entityset=es,
        features=features,
        cutoff_time=pd.Timestamp("12/31/2017"),
    )
    fm = to_pandas(fm)

    for column in data.keys():
        pd.testing.assert_series_equal(
            fm[column],
            answer[column],
            check_names=False,
            check_index=False,
        )

    # cutoff time after all rows, but where clause filters all rows
    if isinstance(parent_df, pd.DataFrame):
        features = [count_where, trend_where, n_most_common_where]
        data = {
            count_where.get_name(): pd.Series([0], dtype="Int64"),
            trend_where.get_name(): pd.Series([np.nan], dtype="float"),
        }
        for name in n_most_common_where.get_feature_names():
            data[name] = pd.Series([np.nan], dtype="category")
    else:
        features = [count_where]
        data = {count_where.get_name(): pd.Series([0], dtype="Int64")}
    answer = pd.DataFrame(data)

    fm2 = calculate_feature_matrix(
        entityset=es,
        features=features,
        cutoff_time=pd.Timestamp("1/4/2018"),
    )
    fm2 = to_pandas(fm2)

    for column in data.keys():
        pd.testing.assert_series_equal(
            fm[column],
            answer[column],
            check_names=False,
            check_index=False,
        )


def test_with_features_built_from_es_metadata(es):
    metadata = es.metadata

    agg_feat = Feature(
        metadata["log"].ww["id"],
        parent_dataframe_name="customers",
        primitive=Count,
    )

    feature_set = FeatureSet([agg_feat])
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)
    df = calculator.run(np.array([0]))
    df = to_pandas(df, index="id")
    v = df[agg_feat.get_name()].values[0]
    assert v == 10


# TODO: Fails with Dask and Spark (conflicting aggregation primitives)
def test_handles_primitive_function_name_uniqueness(es):
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail(
            "Fails with Dask and Spark due conflicting aggregation primitive names",
        )

    class SumTimesN(AggregationPrimitive):
        name = "sum_times_n"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})

        def __init__(self, n):
            self.n = n

        def get_function(self, agg_type="pandas"):
            def my_function(values):
                return values.sum() * self.n

            return my_function

    # works as expected
    f1 = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=SumTimesN(n=1),
    )
    fm = calculate_feature_matrix(features=[f1], entityset=es)

    value_sum = pd.Series([56, 26, 0])
    assert all(fm[f1.get_name()].sort_index() == value_sum)

    # works as expected
    f2 = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=SumTimesN(n=2),
    )
    fm = calculate_feature_matrix(features=[f2], entityset=es)

    double_value_sum = pd.Series([112, 52, 0])
    assert all(fm[f2.get_name()].sort_index() == double_value_sum)

    # same primitive, same column, different args
    fm = calculate_feature_matrix(features=[f1, f2], entityset=es)

    assert all(fm[f1.get_name()].sort_index() == value_sum)
    assert all(fm[f2.get_name()].sort_index() == double_value_sum)

    # different primitives, same function returned by get_function,
    # different base features
    f3 = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Sum,
    )
    f4 = Feature(
        es["log"].ww["purchased"],
        parent_dataframe_name="customers",
        primitive=NumTrue,
    )
    fm = calculate_feature_matrix(features=[f3, f4], entityset=es)

    purchased_sum = pd.Series([10, 1, 1])
    assert all(fm[f3.get_name()].sort_index() == value_sum)
    assert all(fm[f4.get_name()].sort_index() == purchased_sum)

    # different primitives, same function returned by get_function,
    # same base feature
    class Sum1(AggregationPrimitive):
        """Sums elements of a numeric or boolean feature."""

        name = "sum1"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        stack_on_self = False
        stack_on_exclude = [Count]
        default_value = 0

        def get_function(self, agg_type="pandas"):
            return np.sum

    class Sum2(AggregationPrimitive):
        """Sums elements of a numeric or boolean feature."""

        name = "sum2"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        stack_on_self = False
        stack_on_exclude = [Count]
        default_value = 0

        def get_function(self, agg_type="pandas"):
            return np.sum

    class Sum3(AggregationPrimitive):
        """Sums elements of a numeric or boolean feature."""

        name = "sum3"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        stack_on_self = False
        stack_on_exclude = [Count]
        default_value = 0

        def get_function(self, agg_type="pandas"):
            return np.sum

    f5 = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Sum1,
    )
    f6 = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Sum2,
    )
    f7 = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Sum3,
    )
    fm = calculate_feature_matrix(features=[f5, f6, f7], entityset=es)
    assert all(fm[f5.get_name()].sort_index() == value_sum)
    assert all(fm[f6.get_name()].sort_index() == value_sum)
    assert all(fm[f7.get_name()].sort_index() == value_sum)


# No order guarantees w/ Dask
def test_returns_order_of_instance_ids(pd_es):
    feature_set = FeatureSet([Feature(pd_es["customers"].ww["age"])])
    calculator = FeatureSetCalculator(pd_es, time_last=None, feature_set=feature_set)

    instance_ids = [0, 1, 2]
    assert list(pd_es["customers"]["id"]) != instance_ids

    df = calculator.run(np.array(instance_ids))

    assert list(df.index) == instance_ids


def test_calls_progress_callback(es):
    # call with all feature types. make sure progress callback calls sum to 1
    identity = Feature(es["customers"].ww["age"])
    direct = Feature(es["cohorts"].ww["cohort_name"], "customers")
    agg = Feature(
        es["sessions"].ww["id"],
        parent_dataframe_name="customers",
        primitive=Count,
    )
    agg_apply = Feature(
        es["log"].ww["datetime"],
        parent_dataframe_name="customers",
        primitive=TimeSinceLast,
    )  # this feature is handle differently than simple features
    trans = Feature(agg, primitive=Negate)
    trans_full = Feature(agg, primitive=CumSum)
    groupby_trans = Feature(
        agg,
        primitive=CumSum,
        groupby=Feature(es["customers"].ww["cohort"]),
    )

    if es.dataframe_type != Library.PANDAS:
        all_features = [identity, direct, agg, trans]
    else:
        all_features = [
            identity,
            direct,
            agg,
            agg_apply,
            trans,
            trans_full,
            groupby_trans,
        ]

    feature_set = FeatureSet(all_features)
    calculator = FeatureSetCalculator(es, time_last=None, feature_set=feature_set)

    class MockProgressCallback:
        def __init__(self):
            self.total = 0

        def __call__(self, update):
            self.total += update

    mock_progress_callback = MockProgressCallback()

    instance_ids = [0, 1, 2]
    calculator.run(np.array(instance_ids), mock_progress_callback)

    assert np.isclose(mock_progress_callback.total, 1)

    # testing again with a time_last with no data
    feature_set = FeatureSet(all_features)
    calculator = FeatureSetCalculator(
        es,
        time_last=pd.Timestamp("1950"),
        feature_set=feature_set,
    )

    mock_progress_callback = MockProgressCallback()
    calculator.run(np.array(instance_ids), mock_progress_callback)

    assert np.isclose(mock_progress_callback.total, 1)


# precalculated_features is only used with approximate
def test_precalculated_features(pd_es):
    error_msg = (
        "This primitive should never be used because the features are precalculated"
    )

    class ErrorPrim(AggregationPrimitive):
        """A primitive whose function raises an error."""

        name = "error_prim"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})

        def get_function(self, agg_type="pandas"):
            def error(s):
                raise RuntimeError(error_msg)

            return error

    value = Feature(pd_es["log"].ww["value"])
    agg = Feature(value, parent_dataframe_name="sessions", primitive=ErrorPrim)
    agg2 = Feature(agg, parent_dataframe_name="customers", primitive=ErrorPrim)
    direct = Feature(agg2, dataframe_name="sessions")

    # Set up a FeatureSet which knows which features are precalculated.
    precalculated_feature_trie = Trie(default=set, path_constructor=RelationshipPath)
    precalculated_feature_trie.get_node(direct.relationship_path).value.add(
        agg2.unique_name(),
    )
    feature_set = FeatureSet(
        [direct],
        approximate_feature_trie=precalculated_feature_trie,
    )

    # Fake precalculated data.
    values = [0, 1, 2]
    parent_fm = pd.DataFrame({agg2.get_name(): values})
    precalculated_fm_trie = Trie(path_constructor=RelationshipPath)
    precalculated_fm_trie.get_node(direct.relationship_path).value = parent_fm

    calculator = FeatureSetCalculator(
        pd_es,
        feature_set=feature_set,
        precalculated_features=precalculated_fm_trie,
    )

    instance_ids = [0, 2, 3, 5]
    fm = calculator.run(np.array(instance_ids))

    assert list(fm[direct.get_name()]) == [values[0], values[0], values[1], values[2]]

    # Calculating without precalculated features should error.
    with pytest.raises(RuntimeError, match=error_msg):
        FeatureSetCalculator(pd_es, feature_set=FeatureSet([direct])).run(instance_ids)



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import LatLong

from featuretools.primitives.base import TransformPrimitive


class Longitude(TransformPrimitive):
    """Returns the second tuple value in a list of LatLong tuples.
       For use with the LatLong logical type.

    Examples:
        >>> longitude = Longitude()
        >>> longitude([(42.4, -71.1),
        ...            (40.0, -122.4),
        ...            (41.2, -96.75)]).tolist()
        [-71.1, -122.4, -96.75]
    """

    name = "longitude"
    input_types = [ColumnSchema(logical_type=LatLong)]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    description_template = "the longitude of {}"

    def get_function(self):
        def longitude(latlong):
            latlong = np.array(latlong.tolist())
            return latlong[:, 1]

        return longitude



from datetime import datetime
from inspect import isclass
from math import isnan

import numpy as np
import pandas as pd
import pytest
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime

from featuretools import (
    AggregationFeature,
    Feature,
    IdentityFeature,
    Timedelta,
    calculate_feature_matrix,
    dfs,
    primitives,
)
from featuretools.entityset.relationship import RelationshipPath
from featuretools.feature_base.cache import feature_cache
from featuretools.primitives import (
    Count,
    Max,
    Mean,
    Median,
    NMostCommon,
    NumTrue,
    NumUnique,
    Sum,
    TimeSinceFirst,
    TimeSinceLast,
    get_aggregation_primitives,
)
from featuretools.primitives.base import AggregationPrimitive
from featuretools.synthesis.deep_feature_synthesis import DeepFeatureSynthesis, match
from featuretools.tests.testing_utils import backward_path, feature_with_name, to_pandas
from featuretools.utils.gen_utils import Library


@pytest.fixture(autouse=True)
def reset_dfs_cache():
    feature_cache.enabled = False
    feature_cache.clear_all()


def test_get_depth(es):
    log_id_feat = IdentityFeature(es["log"].ww["id"])
    customer_id_feat = IdentityFeature(es["customers"].ww["id"])
    count_logs = Feature(log_id_feat, parent_dataframe_name="sessions", primitive=Count)
    sum_count_logs = Feature(
        count_logs,
        parent_dataframe_name="customers",
        primitive=Sum,
    )
    num_logs_greater_than_5 = sum_count_logs > 5
    count_customers = Feature(
        customer_id_feat,
        parent_dataframe_name="régions",
        where=num_logs_greater_than_5,
        primitive=Count,
    )
    num_customers_region = Feature(count_customers, dataframe_name="customers")

    depth = num_customers_region.get_depth()
    assert depth == 5


def test_makes_count(es):
    dfs = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Count],
        trans_primitives=[],
    )

    features = dfs.build_features()
    assert feature_with_name(features, "device_type")
    assert feature_with_name(features, "customer_id")
    assert feature_with_name(features, "customers.région_id")
    assert feature_with_name(features, "customers.age")
    assert feature_with_name(features, "COUNT(log)")
    assert feature_with_name(features, "customers.COUNT(sessions)")
    assert feature_with_name(features, "customers.régions.language")
    assert feature_with_name(features, "customers.COUNT(log)")


def test_count_null(pd_es):
    class Count(AggregationPrimitive):
        name = "count"
        input_types = [[ColumnSchema(semantic_tags={"foreign_key"})], [ColumnSchema()]]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        stack_on_self = False

        def __init__(self, count_null=True):
            self.count_null = count_null

        def get_function(self):
            def count_func(values):
                if self.count_null:
                    values = values.fillna(0)

                return values.count()

            return count_func

        def generate_name(
            self,
            base_feature_names,
            relationship_path_name,
            parent_dataframe_name,
            where_str,
            use_prev_str,
        ):
            return "COUNT(%s%s%s)" % (relationship_path_name, where_str, use_prev_str)

    count_null = Feature(
        pd_es["log"].ww["value"],
        parent_dataframe_name="sessions",
        primitive=Count(count_null=True),
    )
    feature_matrix = calculate_feature_matrix([count_null], entityset=pd_es)
    values = [5, 4, 1, 2, 3, 2]
    assert (values == feature_matrix[count_null.get_name()]).all()


def test_check_input_types(es):
    count = Feature(
        es["sessions"].ww["id"],
        parent_dataframe_name="customers",
        primitive=Count,
    )
    mean = Feature(count, parent_dataframe_name="régions", primitive=Mean)
    assert mean._check_input_types()

    boolean = count > 3
    mean = Feature(
        count,
        parent_dataframe_name="régions",
        where=boolean,
        primitive=Mean,
    )
    assert mean._check_input_types()


def test_mean_nan(es):
    array = pd.Series([5, 5, 5, 5, 5])
    mean_func_nans_default = Mean().get_function()
    mean_func_nans_false = Mean(skipna=False).get_function()
    mean_func_nans_true = Mean(skipna=True).get_function()
    assert mean_func_nans_default(array) == 5
    assert mean_func_nans_false(array) == 5
    assert mean_func_nans_true(array) == 5
    array = pd.Series([5, np.nan, np.nan, np.nan, np.nan, 10])
    assert mean_func_nans_default(array) == 7.5
    assert isnan(mean_func_nans_false(array))
    assert mean_func_nans_true(array) == 7.5
    array_nans = pd.Series([np.nan, np.nan, np.nan, np.nan])
    assert isnan(mean_func_nans_default(array_nans))
    assert isnan(mean_func_nans_false(array_nans))
    assert isnan(mean_func_nans_true(array_nans))

    # test naming
    default_feat = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Mean,
    )
    assert default_feat.get_name() == "MEAN(log.value)"
    ignore_nan_feat = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Mean(skipna=True),
    )
    assert ignore_nan_feat.get_name() == "MEAN(log.value)"
    include_nan_feat = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Mean(skipna=False),
    )
    assert include_nan_feat.get_name() == "MEAN(log.value, skipna=False)"


def test_init_and_name(es):
    log = es["log"]

    # Add a BooleanNullable column so primitives with that input type get tested
    boolean_nullable = log.ww["purchased"]
    boolean_nullable = boolean_nullable.ww.set_logical_type("BooleanNullable")
    log.ww["boolean_nullable"] = boolean_nullable

    features = [Feature(es["log"].ww[col]) for col in log.columns]

    # check all primitives have name
    for attribute_string in dir(primitives):
        attr = getattr(primitives, attribute_string)
        if isclass(attr):
            if issubclass(attr, AggregationPrimitive) and attr != AggregationPrimitive:
                assert getattr(attr, "name") is not None

    agg_primitives = get_aggregation_primitives().values()
    # If Dask EntitySet use only Dask compatible primitives
    if es.dataframe_type == Library.DASK:
        agg_primitives = [
            prim for prim in agg_primitives if Library.DASK in prim.compatibility
        ]
    if es.dataframe_type == Library.SPARK:
        agg_primitives = [
            prim for prim in agg_primitives if Library.SPARK in prim.compatibility
        ]

    for agg_prim in agg_primitives:
        input_types = agg_prim.input_types
        if not isinstance(input_types[0], list):
            input_types = [input_types]

        # test each allowed input_types for this primitive
        for it in input_types:
            # use the input_types matching function from DFS
            matching_types = match(it, features)
            if len(matching_types) == 0:
                raise Exception("Agg Primitive %s not tested" % agg_prim.name)
            for t in matching_types:
                instance = Feature(
                    t,
                    parent_dataframe_name="sessions",
                    primitive=agg_prim,
                )

                # try to get name and calculate
                instance.get_name()
                calculate_feature_matrix([instance], entityset=es)


def test_invalid_init_args(diamond_es):
    error_text = "parent_dataframe must match first relationship in path"
    with pytest.raises(AssertionError, match=error_text):
        path = backward_path(diamond_es, ["stores", "transactions"])
        AggregationFeature(
            IdentityFeature(diamond_es["transactions"].ww["amount"]),
            "customers",
            Mean,
            relationship_path=path,
        )

    error_text = (
        "Base feature must be defined on the dataframe at the end of relationship_path"
    )
    with pytest.raises(AssertionError, match=error_text):
        path = backward_path(diamond_es, ["regions", "stores"])
        AggregationFeature(
            IdentityFeature(diamond_es["transactions"].ww["amount"]),
            "regions",
            Mean,
            relationship_path=path,
        )

    error_text = "All relationships in path must be backward"
    with pytest.raises(AssertionError, match=error_text):
        backward = backward_path(diamond_es, ["customers", "transactions"])
        forward = RelationshipPath([(True, r) for _, r in backward])
        path = RelationshipPath(list(forward) + list(backward))
        AggregationFeature(
            IdentityFeature(diamond_es["transactions"].ww["amount"]),
            "transactions",
            Mean,
            relationship_path=path,
        )


def test_init_with_multiple_possible_paths(diamond_es):
    error_text = (
        "There are multiple possible paths to the base dataframe. "
        "You must specify a relationship path."
    )
    with pytest.raises(RuntimeError, match=error_text):
        AggregationFeature(
            IdentityFeature(diamond_es["transactions"].ww["amount"]),
            "regions",
            Mean,
        )

    # Does not raise if path specified.
    path = backward_path(diamond_es, ["regions", "customers", "transactions"])
    AggregationFeature(
        IdentityFeature(diamond_es["transactions"].ww["amount"]),
        "regions",
        Mean,
        relationship_path=path,
    )


def test_init_with_single_possible_path(diamond_es):
    # This uses diamond_es to test that there being a cycle somewhere in the
    # graph doesn't cause an error.
    feat = AggregationFeature(
        IdentityFeature(diamond_es["transactions"].ww["amount"]),
        "customers",
        Mean,
    )
    expected_path = backward_path(diamond_es, ["customers", "transactions"])
    assert feat.relationship_path == expected_path


def test_init_with_no_path(diamond_es):
    error_text = 'No backward path from "transactions" to "customers" found.'
    with pytest.raises(RuntimeError, match=error_text):
        AggregationFeature(
            IdentityFeature(diamond_es["customers"].ww["name"]),
            "transactions",
            Count,
        )

    error_text = 'No backward path from "transactions" to "transactions" found.'
    with pytest.raises(RuntimeError, match=error_text):
        AggregationFeature(
            IdentityFeature(diamond_es["transactions"].ww["amount"]),
            "transactions",
            Mean,
        )


def test_name_with_multiple_possible_paths(diamond_es):
    path = backward_path(diamond_es, ["regions", "customers", "transactions"])
    feat = AggregationFeature(
        IdentityFeature(diamond_es["transactions"].ww["amount"]),
        "regions",
        Mean,
        relationship_path=path,
    )

    assert feat.get_name() == "MEAN(customers.transactions.amount)"
    assert feat.relationship_path_name() == "customers.transactions"


def test_copy(games_es):
    home_games = next(
        r for r in games_es.relationships if r._child_column_name == "home_team_id"
    )
    path = RelationshipPath([(False, home_games)])
    feat = AggregationFeature(
        IdentityFeature(games_es["games"].ww["home_team_score"]),
        "teams",
        relationship_path=path,
        primitive=Mean,
    )
    copied = feat.copy()
    assert copied.dataframe_name == feat.dataframe_name
    assert copied.base_features == feat.base_features
    assert copied.relationship_path == feat.relationship_path
    assert copied.primitive == feat.primitive


def test_serialization(es):
    value = IdentityFeature(es["log"].ww["value"])
    primitive = Max()
    max1 = AggregationFeature(value, "customers", primitive)

    path = next(es.find_backward_paths("customers", "log"))
    dictionary = {
        "name": max1.get_name(),
        "base_features": [value.unique_name()],
        "relationship_path": [r.to_dictionary() for r in path],
        "primitive": primitive,
        "where": None,
        "use_previous": None,
    }

    assert dictionary == max1.get_arguments()
    deserialized = AggregationFeature.from_dictionary(
        dictionary,
        es,
        {value.unique_name(): value},
        primitive,
    )
    _assert_agg_feats_equal(max1, deserialized)

    is_purchased = IdentityFeature(es["log"].ww["purchased"])
    use_previous = Timedelta(3, "d")
    max2 = AggregationFeature(
        value,
        "customers",
        primitive,
        where=is_purchased,
        use_previous=use_previous,
    )

    dictionary = {
        "name": max2.get_name(),
        "base_features": [value.unique_name()],
        "relationship_path": [r.to_dictionary() for r in path],
        "primitive": primitive,
        "where": is_purchased.unique_name(),
        "use_previous": use_previous.get_arguments(),
    }

    assert dictionary == max2.get_arguments()
    dependencies = {
        value.unique_name(): value,
        is_purchased.unique_name(): is_purchased,
    }
    deserialized = AggregationFeature.from_dictionary(
        dictionary,
        es,
        dependencies,
        primitive,
    )
    _assert_agg_feats_equal(max2, deserialized)


def test_time_since_last(pd_es):
    f = Feature(
        pd_es["log"].ww["datetime"],
        parent_dataframe_name="customers",
        primitive=TimeSinceLast,
    )
    fm = calculate_feature_matrix(
        [f],
        entityset=pd_es,
        instance_ids=[0, 1, 2],
        cutoff_time=datetime(2015, 6, 8),
    )

    correct = [131376000.0, 131289534.0, 131287797.0]
    # note: must round to nearest second
    assert all(fm[f.get_name()].round().values == correct)


def test_time_since_first(pd_es):
    f = Feature(
        pd_es["log"].ww["datetime"],
        parent_dataframe_name="customers",
        primitive=TimeSinceFirst,
    )
    fm = calculate_feature_matrix(
        [f],
        entityset=pd_es,
        instance_ids=[0, 1, 2],
        cutoff_time=datetime(2015, 6, 8),
    )

    correct = [131376600.0, 131289600.0, 131287800.0]
    # note: must round to nearest second
    assert all(fm[f.get_name()].round().values == correct)


def test_median(pd_es):
    f = Feature(
        pd_es["log"].ww["value_many_nans"],
        parent_dataframe_name="customers",
        primitive=Median,
    )
    fm = calculate_feature_matrix(
        [f],
        entityset=pd_es,
        instance_ids=[0, 1, 2],
        cutoff_time=datetime(2015, 6, 8),
    )

    correct = [1, 3, np.nan]
    np.testing.assert_equal(fm[f.get_name()].values, correct)


def test_agg_same_method_name(es):
    """
    Pandas relies on the function name when calculating aggregations. This means if a two
    primitives with the same function name are applied to the same column, pandas
    can't differentiate them. We have a work around to this based on the name property
    that we test here.
    """
    # TODO: Update to work with Dask and Spark
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail("Need to update to work with Dask and Spark EntitySets")

    # test with normally defined functions
    class Sum(AggregationPrimitive):
        name = "sum"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})

        def get_function(self):
            def custom_primitive(x):
                return x.sum()

            return custom_primitive

    class Max(AggregationPrimitive):
        name = "max"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})

        def get_function(self):
            def custom_primitive(x):
                return x.max()

            return custom_primitive

    f_sum = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Sum,
    )
    f_max = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Max,
    )

    fm = calculate_feature_matrix([f_sum, f_max], entityset=es)
    assert fm.columns.tolist() == [f_sum.get_name(), f_max.get_name()]

    # test with lambdas
    class Sum(AggregationPrimitive):
        name = "sum"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})

        def get_function(self):
            return lambda x: x.sum()

    class Max(AggregationPrimitive):
        name = "max"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})

        def get_function(self):
            return lambda x: x.max()

    f_sum = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Sum,
    )
    f_max = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Max,
    )
    fm = calculate_feature_matrix([f_sum, f_max], entityset=es)
    assert fm.columns.tolist() == [f_sum.get_name(), f_max.get_name()]


def test_time_since_last_custom(pd_es):
    class TimeSinceLast(AggregationPrimitive):
        name = "time_since_last"
        input_types = [
            ColumnSchema(logical_type=Datetime, semantic_tags={"time_index"}),
        ]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        uses_calc_time = True

        def get_function(self):
            def time_since_last(values, time):
                time_since = time - values.iloc[0]
                return time_since.total_seconds()

            return time_since_last

    f = Feature(
        pd_es["log"].ww["datetime"],
        parent_dataframe_name="customers",
        primitive=TimeSinceLast,
    )
    fm = calculate_feature_matrix(
        [f],
        entityset=pd_es,
        instance_ids=[0, 1, 2],
        cutoff_time=datetime(2015, 6, 8),
    )

    correct = [131376600, 131289600, 131287800]
    # note: must round to nearest second
    assert all(fm[f.get_name()].round().values == correct)


def test_custom_primitive_multiple_inputs(pd_es):
    class MeanSunday(AggregationPrimitive):
        name = "mean_sunday"
        input_types = [
            ColumnSchema(semantic_tags={"numeric"}),
            ColumnSchema(logical_type=Datetime),
        ]
        return_type = ColumnSchema(semantic_tags={"numeric"})

        def get_function(self):
            def mean_sunday(numeric, datetime):
                """
                Finds the mean of non-null values of a feature that occurred on Sundays
                """
                days = pd.DatetimeIndex(datetime).weekday.values
                df = pd.DataFrame({"numeric": numeric, "time": days})
                return df[df["time"] == 6]["numeric"].mean()

            return mean_sunday

    fm, features = dfs(
        entityset=pd_es,
        target_dataframe_name="sessions",
        agg_primitives=[MeanSunday],
        trans_primitives=[],
    )
    mean_sunday_value = pd.Series([None, None, None, 2.5, 7, None])
    iterator = zip(fm["MEAN_SUNDAY(log.value, datetime)"], mean_sunday_value)
    for x, y in iterator:
        assert (pd.isnull(x) and pd.isnull(y)) or (x == y)

    pd_es.add_interesting_values()
    mean_sunday_value_priority_0 = pd.Series([None, None, None, 2.5, 0, None])
    fm, features = dfs(
        entityset=pd_es,
        target_dataframe_name="sessions",
        agg_primitives=[MeanSunday],
        trans_primitives=[],
        where_primitives=[MeanSunday],
    )
    where_feat = "MEAN_SUNDAY(log.value, datetime WHERE priority_level = 0)"
    for x, y in zip(fm[where_feat], mean_sunday_value_priority_0):
        assert (pd.isnull(x) and pd.isnull(y)) or (x == y)


def test_custom_primitive_default_kwargs(es):
    class SumNTimes(AggregationPrimitive):
        name = "sum_n_times"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})

        def __init__(self, n=1):
            self.n = n

    sum_n_1_n = 1
    sum_n_1_base_f = Feature(es["log"].ww["value"])
    sum_n_1 = Feature(
        [sum_n_1_base_f],
        parent_dataframe_name="sessions",
        primitive=SumNTimes(n=sum_n_1_n),
    )
    sum_n_2_n = 2
    sum_n_2_base_f = Feature(es["log"].ww["value_2"])
    sum_n_2 = Feature(
        [sum_n_2_base_f],
        parent_dataframe_name="sessions",
        primitive=SumNTimes(n=sum_n_2_n),
    )
    assert sum_n_1_base_f == sum_n_1.base_features[0]
    assert sum_n_1_n == sum_n_1.primitive.n
    assert sum_n_2_base_f == sum_n_2.base_features[0]
    assert sum_n_2_n == sum_n_2.primitive.n


def test_makes_numtrue(es):
    if es.dataframe_type == Library.SPARK:
        pytest.xfail("Spark EntitySets do not support NumTrue primitive")
    dfs = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[NumTrue],
        trans_primitives=[],
    )
    features = dfs.build_features()
    assert feature_with_name(features, "customers.NUM_TRUE(log.purchased)")
    assert feature_with_name(features, "NUM_TRUE(log.purchased)")


def test_make_three_most_common(pd_es):
    class NMostCommoner(AggregationPrimitive):
        name = "pd_top3"
        input_types = ([ColumnSchema(semantic_tags={"category"})],)
        return_type = None
        number_output_features = 3

        def get_function(self):
            def pd_top3(x):
                counts = x.value_counts()
                counts = counts[counts > 0]
                array = np.array(counts[:3].index)
                if len(array) < 3:
                    filler = np.full(3 - len(array), np.nan)
                    array = np.append(array, filler)
                return array

            return pd_top3

    fm, features = dfs(
        entityset=pd_es,
        target_dataframe_name="customers",
        instance_ids=[0, 1, 2],
        agg_primitives=[NMostCommoner],
        trans_primitives=[],
    )

    df = fm[["PD_TOP3(log.product_id)[%s]" % i for i in range(3)]]

    assert set(df.iloc[0].values[:2]) == set(
        ["coke zero", "toothpaste"],
    )  # coke zero and toothpaste have same number of occurrences
    assert df.iloc[0].values[2] in [
        "car",
        "brown bag",
    ]  # so just check that the top two match

    assert (
        df.iloc[1]
        .reset_index(drop=True)
        .equals(pd.Series(["coke zero", "Haribo sugar-free gummy bears", np.nan]))
    )
    assert (
        df.iloc[2]
        .reset_index(drop=True)
        .equals(pd.Series(["taco clock", np.nan, np.nan]))
    )


def test_stacking_multi(pd_es):
    threecommon = NMostCommon(3)
    tc = Feature(
        pd_es["log"].ww["product_id"],
        parent_dataframe_name="sessions",
        primitive=threecommon,
    )

    stacked = []
    for i in range(3):
        stacked.append(
            Feature(tc[i], parent_dataframe_name="customers", primitive=NumUnique),
        )

    fm = calculate_feature_matrix(stacked, entityset=pd_es, instance_ids=[0, 1, 2])

    correct_vals = [[3, 2, 1], [2, 1, 0], [0, 0, 0]]
    correct_vals1 = [[3, 1, 1], [2, 1, 0], [0, 0, 0]]
    # either of the above can be correct, and the outcome depends on the sorting of
    # two values in the initial n most common function, which changes arbitrarily.

    for i in range(3):
        f = "NUM_UNIQUE(sessions.N_MOST_COMMON(log.product_id)[%d])" % i
        cols = fm.columns
        assert f in cols
        assert (
            fm[cols[i]].tolist() == correct_vals[i]
            or fm[cols[i]].tolist() == correct_vals1[i]
        )


def test_use_previous_pd_dateoffset(es):
    total_events_pd = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="customers",
        use_previous=pd.DateOffset(hours=47, minutes=60),
        primitive=Count,
    )

    feature_matrix = calculate_feature_matrix(
        [total_events_pd],
        es,
        cutoff_time=pd.Timestamp("2011-04-11 10:31:30"),
        instance_ids=[0, 1, 2],
    )
    feature_matrix = to_pandas(feature_matrix, index="id", sort_index=True)
    col_name = list(feature_matrix.head().keys())[0]
    assert (feature_matrix[col_name] == [1, 5, 2]).all()


def _assert_agg_feats_equal(f1, f2):
    assert f1.unique_name() == f2.unique_name()
    assert f1.child_dataframe_name == f2.child_dataframe_name
    assert f1.parent_dataframe_name == f2.parent_dataframe_name
    assert f1.relationship_path == f2.relationship_path
    assert f1.use_previous == f2.use_previous


def test_override_multi_feature_names(pd_es):
    def gen_custom_names(
        primitive,
        base_feature_names,
        relationship_path_name,
        parent_dataframe_name,
        where_str,
        use_prev_str,
    ):
        base_string = "Custom_%s({}.{})".format(
            parent_dataframe_name,
            base_feature_names,
        )
        return [base_string % i for i in range(primitive.number_output_features)]

    class NMostCommoner(AggregationPrimitive):
        name = "pd_top3"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"category"})
        number_output_features = 3

        def generate_names(
            self,
            base_feature_names,
            relationship_path_name,
            parent_dataframe_name,
            where_str,
            use_prev_str,
        ):
            return gen_custom_names(
                self,
                base_feature_names,
                relationship_path_name,
                parent_dataframe_name,
                where_str,
                use_prev_str,
            )

    fm, features = dfs(
        entityset=pd_es,
        target_dataframe_name="products",
        instance_ids=[0, 1, 2],
        agg_primitives=[NMostCommoner],
        trans_primitives=[],
    )

    expected_names = []
    base_names = [["value"], ["value_2"], ["value_many_nans"]]
    for name in base_names:
        expected_names += gen_custom_names(
            NMostCommoner,
            name,
            None,
            "products",
            None,
            None,
        )

    for name in expected_names:
        assert name in fm.columns



import numpy as np
from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive
from featuretools.primitives.standard.aggregation.count import Count
from featuretools.utils.gen_utils import Library


class Sum(AggregationPrimitive):
    """Calculates the total addition, ignoring `NaN`.

    Examples:
        >>> sum = Sum()
        >>> sum([1, 2, 3, 4, 5, None])
        15.0
    """

    name = "sum"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    stack_on_self = False
    stack_on_exclude = [Count]
    default_value = 0
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the sum of {}"

    def get_function(self, agg_type=Library.PANDAS):
        if agg_type in [Library.DASK, Library.SPARK]:
            return "sum"

        return np.sum
